#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np

from ..representations.polynomials import chebychev, poly_constraints

# from .. import TFmath
from .rational_attributes import PolyFit
from .rational_bases import SFilterBase
from .rational_algorithms import PolyFilterAlgorithmsIm


# must be in this order for the ZPKsf to be set in the proper order
class ChebychevFilter(SFilterBase, PolyFit, PolyFilterAlgorithmsIm):
    phase_missing = False
    root_constraint = SFilterBase.RBalgo.root_constraints.mirror_imag
    poly = chebychev

    def phi_gain_adj_roots(self, roots):
        if len(roots) % 2:
            return 1j
        else:
            return 1

    def phi_gain_adj_pvec(self, pvec):
        if len(pvec) % 2:
            return 1
        else:
            return 1j

    def phi_rep(self, F_Hz):
        """
        function mapping roots from frequency domain to fit domain
        """
        return F_Hz / self.F_max_Hz

    def phi_rep_native_lnG(self, rB):
        """
        function mapping roots to native domain (S) from fit representation domain
        outputs both the remapped roots, as well as the lnG adjustment
        """
        rB = self.RBalgo.expect(rB, self.RBalgo.root_constraints.mirror_imag)
        lnG = len(rB) * np.log(self.F_max_Hz)
        return rB * (1j * self.F_max_Hz), lnG

    def phi_native_rep_lnG(self, rB):
        """
        function mapping roots from native domain (S) to fit representation domain
        outputs both the remapped roots, as well as the lnG adjustment
        """
        rB = self.RBalgo.expect(rB, self.RBalgo.root_constraints.mirror_real)
        lnG = len(rB) * np.log(self.F_max_Hz)
        return rB * (-1j / self.F_max_Hz), -lnG

    def phi_rep_Snative(self, rB):
        """
        Maps the internal representation to Snative, the output RootBunch can
        have a specialty constraint added.
        """
        rB = self.RBalgo.expect(rB, self.RBalgo.root_constraints.mirror_imag)
        return rB * (1j * self.F_max_Hz)

    def phi_Snative_rep(self, rB):
        """ """
        rB = self.RBalgo.expect(rB, self.RBalgo.root_constraints.mirror_real)
        return rB * (-1j / self.F_max_Hz)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from matplotlib import gridspec
import matplotlib.pyplot as plt

from ..utilities.mpl import generate_ax
from . import utilities
from .utilities import ZPKrep2Sf


def plot_ZP(
    self,
    fitter,
    db_nyquist=True,
    gs_base=gridspec.GridSpec(1, 1)[0],
    fig=None,
    horizontal=True,
    wspacing=0.25,
    hspace=0.2,
    ax_bunch=None,
):
    axB = generate_ax(ax_bunch)

    if fig is None:
        fig = plt.figure()
        if horizontal:
            fig.set_size_inches(12, 5)
        else:
            fig.set_size_inches(5, 12)
    axB.fig = fig

    if horizontal:
        N_horizontal = 3
        N_vertical = 1
    else:
        N_horizontal = 1
        N_vertical = 3

    gs_DC = gridspec.GridSpecFromSubplotSpec(
        N_vertical,
        N_horizontal,
        subplot_spec=gs_base,
        wspace=wspacing,
        hspace=hspace,
        # height_ratios = height_ratio_list,
        # width_ratios = width_ratios,
    )

    if horizontal:
        axB.ax_poles = axB.fig.add_subplot(gs_DC[0, 0], polar=True)
        axB.ax_dual = axB.fig.add_subplot(gs_DC[0, 1], polar=True)
        axB.ax_zeros = axB.fig.add_subplot(gs_DC[0, 2], polar=True)
    else:
        axB.ax_poles = axB.fig.add_subplot(gs_DC[0, 0], polar=True)
        axB.ax_dual = axB.fig.add_subplot(gs_DC[1, 0], polar=True)
        axB.ax_zeros = axB.fig.add_subplot(gs_DC[2, 0], polar=True)
    axB.ax_poles.set_aspect("equal", "box", anchor="C")
    axB.ax_dual.set_aspect("equal", "box", anchor="C")
    axB.ax_zeros.set_aspect("equal", "box", anchor="C")
    axB.ax_poles.set_rlabel_position(225)
    axB.ax_dual.set_rlabel_position(225)
    axB.ax_zeros.set_rlabel_position(225)

    def plot_poles(ax):
        p = fitter.poles_full
        r = abs(p)
        theta = np.angle(p)
        select = abs(p) < 1
        if db_nyquist:
            F_largest_Hz = np.max(fitter.F_Hz)
            r_pos = np.maximum(
                -10
                * np.log10(
                    utilities.BWz(r[select], fitter.F_nyquist_Hz / F_largest_Hz)
                ),
                0,
            )
            r_neg = np.maximum(
                -10
                * np.log10(
                    utilities.BWz_neg(r[~select], fitter.F_nyquist_Hz / F_largest_Hz)
                ),
                0,
            )
        else:
            r_pos = r[select]
            r_neg = r[~select]

        ax.scatter(
            theta[select],
            r_pos,
            s=15,
            facecolor=self.poleID_color,
            linewidths=0.3,
            edgecolors="black",
            label="stable(<1)",
        )
        ax.scatter(
            theta[~select],
            r_neg,
            s=15,
            facecolor=self.poleOD_color,
            linewidths=0.3,
            edgecolors="black",
            label="unstable(>1)",
        )
        return max(np.concatenate([[-10], r_pos, r_neg]))

    def plot_zeros(ax):
        z = fitter.zeros_full
        r = abs(z)
        theta = np.angle(z)
        select = abs(z) < 1
        if db_nyquist:
            F_largest_Hz = np.max(fitter.F_Hz)
            r_pos = np.maximum(
                -10
                * np.log10(
                    utilities.BWz(r[select], fitter.F_nyquist_Hz / F_largest_Hz)
                ),
                0,
            )
            r_neg = np.maximum(
                -10
                * np.log10(
                    utilities.BWz_neg(r[~select], fitter.F_nyquist_Hz / F_largest_Hz)
                ),
                0,
            )
        else:
            r_pos = r[select]
            r_neg = r[~select]

        ax.scatter(
            theta[select],
            r_pos,
            s=15,
            facecolor=self.zeroID_color,
            linewidths=0.5,
            # edgecolors = 'black',
            marker="x",
            label="mindelay (<1)",
        )
        ax.scatter(
            theta[~select],
            r_neg,
            s=15,
            facecolor=self.zeroOD_color,
            linewidths=0.5,
            # edgecolors = 'black',
            marker="x",
            label="delayed (>1)",
        )
        return max(np.concatenate([r_pos, r_neg]))

    r_max = plot_poles(axB.ax_poles)
    plot_poles(axB.ax_dual)
    r_max = max(plot_zeros(axB.ax_dual), r_max)
    plot_zeros(axB.ax_zeros)
    axB.ax_poles.set_title("poles")
    axB.ax_dual.set_title("dual")
    axB.ax_zeros.set_title("zeros")
    axB.ax_poles.set_ylim(0, (r_max // 3 + 1) * 3)
    axB.ax_dual.set_ylim(0, (r_max // 3 + 1) * 3)
    axB.ax_zeros.set_ylim(0, (r_max // 3 + 1) * 3)
    # ax1.axhline(40, color = 'purple')
    # axB.ax_dual.axhline(40, color = 'purple')
    # axB.ax_zeros.axhline(40, color = 'purple')

    num = 9
    axB.ax_poles.set_xticks(np.linspace(0, np.pi, num))
    axB.ax_poles.set_xticklabels(
        ["{0:.1f}".format(f) for f in np.linspace(0, fitter.F_nyquist_Hz, num)]
    )
    axB.ax_dual.set_xticks(np.linspace(0, np.pi, num))
    axB.ax_dual.set_xticklabels(
        ["{0:.1f}".format(f) for f in np.linspace(0, fitter.F_nyquist_Hz, num)]
    )
    axB.ax_zeros.set_xticks(np.linspace(0, np.pi, num))
    axB.ax_zeros.set_xticklabels(
        ["{0:.1f}".format(f) for f in np.linspace(0, fitter.F_nyquist_Hz, num)]
    )
    # c = patches.Circle((0, 0), 1, color = 'black', fill = False)
    # ax.add_artist(c)
    axB.ax_poles.legend(loc="lower right", fontsize=8)
    axB.ax_zeros.legend(loc="lower right", fontsize=8)
    axB.ax_poles.grid(b=True)
    axB.ax_dual.grid(b=True)
    axB.ax_zeros.grid(b=True)

    axB.finalize()
    return axB


def plot_ZP_S(
    self,
    fitter,
    db_nyquist=True,
    gs_base=gridspec.GridSpec(1, 1)[0],
    fig=None,
    horizontal=True,
    wspacing=0.25,
    hspace=0.2,
    ax_bunch=None,
):
    axB = generate_ax(ax_bunch)

    if fig is None:
        fig = plt.figure()
        if horizontal:
            fig.set_size_inches(12, 5)
        else:
            fig.set_size_inches(5, 12)
    axB.fig = fig

    if horizontal:
        N_horizontal = 3
        N_vertical = 1
    else:
        N_horizontal = 1
        N_vertical = 3

    gs_DC = gridspec.GridSpecFromSubplotSpec(
        N_vertical,
        N_horizontal,
        subplot_spec=gs_base,
        wspace=wspacing,
        hspace=hspace,
        # height_ratios = height_ratio_list,
        # width_ratios = width_ratios,
    )

    if horizontal:
        axB.ax_poles = axB.fig.add_subplot(gs_DC[0, 0])
        axB.ax_dual = axB.fig.add_subplot(gs_DC[0, 1])
        axB.ax_zeros = axB.fig.add_subplot(gs_DC[0, 2])
    else:
        axB.ax_poles = axB.fig.add_subplot(gs_DC[0, 0])
        axB.ax_dual = axB.fig.add_subplot(gs_DC[1, 0])
        axB.ax_zeros = axB.fig.add_subplot(gs_DC[2, 0])
    axB.ax_poles.yaxis.tick_right()
    axB.ax_dual.yaxis.tick_right()
    axB.ax_zeros.yaxis.tick_right()

    axB.ax_poles.yaxis.set_label_position("right")
    axB.ax_dual.yaxis.set_label_position("right")
    axB.ax_zeros.yaxis.set_label_position("right")

    def plot_poles(ax, ZPKrep):
        p = ZPKrep.poles.fullplane
        F_Hz = p.imag
        BW_Hz = p.real
        select = p.real < 0

        ax.scatter(
            -BW_Hz[select],
            F_Hz[select],
            s=15,
            facecolor=self.poleID_color,
            linewidths=0.3,
            edgecolors="black",
            label="stable(re<0)",
        )
        ax.scatter(
            BW_Hz[~select],
            F_Hz[~select],
            s=15,
            facecolor=self.poleOD_color,
            linewidths=0.3,
            edgecolors="black",
            label="unstable(re>0)",
        )
        return BW_Hz

    def plot_zeros(ax, ZPKrep):
        z = ZPKrep.zeros.fullplane
        F_Hz = z.imag
        BW_Hz = z.real
        select = z.real < 0

        ax.scatter(
            -BW_Hz[select],
            F_Hz[select],
            s=15,
            facecolor=self.zeroID_color,
            linewidths=0.5,
            # edgecolors = 'black',
            marker="x",
            label="mindelay (re<0)",
        )
        ax.scatter(
            BW_Hz[~select],
            F_Hz[~select],
            s=15,
            facecolor=self.zeroOD_color,
            linewidths=0.5,
            # edgecolors = 'black',
            marker="x",
            label="delayed (re>0)",
        )
        return BW_Hz

    def x_limits(*r_vals):
        r_vals = np.concatenate(r_vals)
        if len(r_vals) == 0:
            return 1, 0.1
        rmin = min(abs(r_vals))
        rmax = max(abs(r_vals))
        return rmax * 3, rmin / 3

    ZPKrep = ZPKrep2Sf(fitter.ZPKrep)
    r_p = plot_poles(axB.ax_poles, ZPKrep)
    plot_poles(axB.ax_dual, ZPKrep)
    r_z = plot_zeros(axB.ax_dual, ZPKrep)
    plot_zeros(axB.ax_zeros, ZPKrep)
    axB.ax_poles.set_ylabel("poles Freq [Hz]")
    axB.ax_dual.set_ylabel("dual Freq [Hz]")
    axB.ax_zeros.set_ylabel("zeros Freq [Hz]")

    if len(r_p) > 0:
        axB.ax_poles.set_xscale("log_zoom")

    axB.ax_poles.set_xlim(x_limits(r_p))
    if len(r_p) > 0 or len(r_z) > 0:
        axB.ax_dual.set_xscale("log_zoom")
    axB.ax_dual.set_xlim(x_limits(r_p, r_z))
    if len(r_z) > 0:
        axB.ax_zeros.set_xscale("log_zoom")
    axB.ax_zeros.set_xlim(x_limits(r_z))

    axB.ax_poles.legend(loc="lower left", fontsize=8)
    axB.ax_zeros.legend(loc="lower left", fontsize=8)
    axB.ax_poles.grid(b=True)
    axB.ax_dual.grid(b=True)
    axB.ax_zeros.grid(b=True)

    axB.finalize()
    return axB



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


def configuration(parent_package="", top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration(
        "IIRrational", parent_package=parent_package, top_path=top_path
    )
    # config.set_options(
    #    ignore_setup_xxx_py=True,
    #    assume_default_configuration=True,
    #    delegate_options_to_subpackages=True,
    #    quiet=True
    # )

    config.add_subpackage("math")
    return config


if __name__ == "__main__":
    from setuptools import setup
    from numpy.distutils.core import setup

    setup(
        name="IIRrational",
        version="1.0.0.dev1",
        description="Fast Transfer Function Fitting",
        author="Lee McCuller",
        author_email="Lee.McCuller@gmail.com",
        py_modules=[
            "IIRrational",
        ],
        configuration=configuration,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from ..codings_cmn import (
    CodingType,
    # Ipi,
    # I2pi
)

# import scipy.linalg


class CodingDelayPairNl(CodingType):
    N_parameters = 2
    unstable = False
    p_nlF_Hz = 0
    p_nl_BW = 0
    min_BW_Hz = 0

    def setup(self):
        self.N_parameters = 2

    def update(self, A, B):
        nlF_Hz, nl_BW = A, B
        self.p_nlF_Hz = nlF_Hz
        self.p_nl_BW = nl_BW

    def reduce(self):
        return [
            self.p_nlF_Hz,
            self.p_nl_BW,
        ]

    def check_dist_limit(self, F_Hz=None, thresh=1):
        if self.sys.distance_limit_auto >= thresh:
            if F_Hz is None:
                self.min_BW_Hz, D = self.sys.distance_limit(
                    self.F_Hz, with_derivative=True
                )
            else:
                self.min_BW_Hz, D = self.sys.distance_limit(F_Hz, with_derivative=True)
            return D
        return 0

    @property
    def F_Hz(self):
        return self.p_nlF_Hz ** 2

    @F_Hz.setter
    def F_Hz(self, val):
        self.p_nlF_Hz = val ** 0.5

    @property
    def BW_Hz(self):
        return self.min_BW_Hz + self.p_nl_BW ** 2

    @BW_Hz.setter
    def BW_Hz(self, val):
        if val > self.min_BW_Hz:
            self.p_nl_BW = (val - self.min_BW_Hz) ** 0.5
        else:
            self.p_nl_BW = self.min_BW_Hz ** 0.5

    def transfer(self):
        self.check_dist_limit(thresh=2)
        # frequency, logarithmic amplitude
        if not self.unstable:
            r = -self.BW_Hz
        else:
            r = self.BW_Hz
        i = self.F_Hz
        X = self.sys.Xsf_grid
        Xsq = self.sys.Xsf_grid_sq
        xterm = (r * r + i * i) + Xsq
        return (xterm + 2 * r * X) / (xterm - 2 * r * X)

    def derivative(self):
        # real/imaginary part of root
        if not self.unstable:
            r = -self.BW_Hz
            DrDp = -2 * self.p_nl_BW
        else:
            r = self.BW_Hz
            DrDp = 2 * self.p_nl_BW

        i = self.F_Hz
        DiDp = 2 * self.p_nlF_Hz

        X = self.sys.Xsf_grid

        Xsq = self.sys.Xsf_grid_sq
        xterm = (r * r + i * i) + Xsq
        xfer = (xterm + 2 * r * X) / (xterm - 2 * r * X)

        return [
            DiDp * (2 * i) * xfer * (1 / (xterm + 2 * r * X) - 1 / (xterm - 2 * r * X)),
            DrDp
            * xfer
            * (
                (2 * r + 2 * X) / (xterm + 2 * r * X)
                - (2 * r - 2 * X) / (xterm - 2 * r * X)
            ),
        ]


class CodingDelayPair(CodingType):
    N_parameters = 2
    p_nlF_Hz = 0
    p_BW = 0

    def setup(self):
        self.N_parameters = 2

    def update(self, A, B):
        nlF_Hz, nl_BW = A, B
        self.p_nlF_Hz = nlF_Hz
        self.p_BW = nl_BW

    def reduce(self):
        return [
            self.p_nlF_Hz,
            self.p_BW,
        ]

    @property
    def F_Hz(self):
        return self.p_nlF_Hz ** 2

    @F_Hz.setter
    def F_Hz(self, val):
        self.p_nlF_Hz = val ** 0.5

    @property
    def BW_Hz(self):
        return self.p_BW

    @BW_Hz.setter
    def BW_Hz(self, val):
        self.p_BW = val

    def transfer(self):
        # frequency, logarithmic amplitude
        r = -self.BW_Hz
        i = self.F_Hz
        X = self.sys.Xsf_grid
        Xsq = self.sys.Xsf_grid_sq
        xterm = (r * r + i * i) + Xsq
        return (xterm + 2 * r * X) / (xterm - 2 * r * X)

    def derivative(self):
        # real/imaginary part of root
        r = -self.BW_Hz
        DrDp = -1

        i = self.F_Hz
        DiDp = 2 * self.p_nlF_Hz

        X = self.sys.Xsf_grid

        Xsq = self.sys.Xsf_grid_sq
        xterm = (r * r + i * i) + Xsq
        xfer = (xterm + 2 * r * X) / (xterm - 2 * r * X)

        return [
            DiDp * (2 * i) * xfer * (1 / (xterm + 2 * r * X) - 1 / (xterm - 2 * r * X)),
            DrDp
            * xfer
            * (
                (2 * r + 2 * X) / (xterm + 2 * r * X)
                - (2 * r - 2 * X) / (xterm - 2 * r * X)
            ),
        ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import argparse
import os
import sys
import numpy as np
import collections
from wield import declarative
from wield.bunch import Bunch, DeepBunch

from .. import file_io

from .. import version as IIRversion
#from .. import auto_version as IIRversion_auto
from ..utilities.strings import padding_remove
from .. import fitters_ZPK

from . import arguments
from .arguments import pyargparse
from .arguments import cmdline
from . import fit_aid

from .data2filter import data2filter

# from .arguments import ArgumentError


class DataError(ValueError):
    pass


def find_groups(fdict, groups):
    gdict = dict()
    used_groups = []
    for key in reversed(groups):
        try:
            subgroup = file_io.subkey_search(fdict, key, default=declarative.NOARG)
        except TypeError:
            continue
        if subgroup is not declarative.NOARG:
            used_groups.append(used_groups)
            gdict.update(subgroup)
    return gdict, used_groups


def dump_obj(d):
    if isinstance(d, np.ndarray):
        if d.shape == ():
            return "[]"
        if len(d) < 8:
            return "[" + (", ".join(dump_obj(o) for o in d)) + "]"
        else:
            return "[{} {} array]".format(d.shape, d.dtype)
    elif isinstance(d, list):
        d = np.asarray(d)
        if len(d) < 8:
            return "[" + (", ".join(dump_obj(o) for o in d)) + "]"
        else:
            return "[{} {} list]".format(len(d), d.dtype)
    elif isinstance(d, tuple):
        d = np.asarray(d)
        if len(d) < 8:
            return "(" + (", ".join(dump_obj(o) for o in d)) + ")"
        else:
            return "[{} {} tuple]".format(len(d), d.dtype)
    else:
        return str(d)


def dump_fdict_keys(d, k=None, depth=-1, file=sys.stdout):
    subprint = False
    if isinstance(d, collections.Mapping):
        if k is not None:
            print("{}{}:".format("  " * depth, k), file=file)
        for subk, v in sorted(d.items()):
            dump_fdict_keys(v, subk, depth=depth + 1)
    elif isinstance(d, (np.ndarray, list, tuple)):
        d = np.asarray(d)
        if d.shape == ():
            return dump_fdict_keys(d.item(), k, depth, file=file)
        else:
            print("{}{}: {}".format("  " * depth, k, dump_obj(d)), file=file)
        if d.dtype == object and len(d) < 8:
            subprint = True
    else:
        print("{}{}: {}".format("  " * depth, k, str(d)), file=file)

    if subprint:
        for idx, v in enumerate(d):
            dump_fdict_keys(v, idx, depth=depth + 1)


def find_elements(fdict, groups, allow_dicts):
    used_keys = []
    elements = []
    for key in reversed(groups):
        try:
            val = file_io.subkey_search(fdict, key)
        except TypeError:
            continue
        except KeyError:
            continue

        if not allow_dicts and isinstance(val, collections.Mapping):
            continue

        used_keys.append(used_keys)
        elements.append(val)
    return elements, used_keys


def config_load(aid, args, kwfull, mode):
    # CONFIG File Loading
    cfiles = arguments.grab_kwargs(aid, args, cmdline.kw_hints, "config")
    cgroup = arguments.grab_kwargs(aid, args, cmdline.kw_hints, "config_group")

    fdicts_conf = dict()
    for cfname in reversed(cfiles):
        try:
            fB_conf = file_io.determine_type(cfname)
        except KeyError:
            raise arguments.ArgumentError(
                ("Unrecognized file type for {}").format(cfname)
            )
        if fB_conf.ftype == "csv":
            raise arguments.ArgumentError("Configuration files may not be csv")
        elif fB_conf.ftype == "ini":
            raise arguments.ArgumentError("ini files not supported")
        cgroup_use = cgroup
        if fB_conf.subkey is not None:
            cgroup_use = cmdline.mapcheck_cslist(fB_conf.subkey)

        fdict_conf = file_io.load_any(
            fname=fB_conf.fname,
            ftype=fB_conf.ftype,
        )
        fdicts_conf[cfname] = fdict_conf
        # detect and warn if more than one are found
        conf_dict, used_cgroups = find_groups(fdict_conf, cgroup_use)
        if len(used_cgroups) == 0:
            print(
                ("No config groups found in config file {}").format(fB_conf.fname),
                file=sys.stderr,
            )
            print("The config file layout is:", file=sys.stderr)
            dump_fdict_keys(fdict_conf, file=sys.stderr)
            # TODO dump data file layout
            raise SystemExit(1)
        elif len(used_cgroups) > 1:
            print(
                (
                    "Warning: multiple possible config groups found in config file {}"
                ).format(fB_conf.fname),
                file=sys.stderr,
            )
            print("Groups aggregated: {}".format(used_cgroups), file=sys.stderr)

        if mode == "printconfig":
            print("config file layout for mode=printconfig")
            print("file: {}".fortmat(fB_conf.fname))
            dump_fdict_keys(fdict_conf)

        kwfull.update(conf_dict)

    kwfull.update(args)
    return Bunch(
        cfiles=cfiles,
        fdicts=fdicts_conf,
        cgroup=cgroup,
    )


def data_load(aid, args, kwfull, mode, confB):
    # DATA FILE LOADING
    dfile = args["datafile"]
    try:
        fB_data = file_io.determine_type(dfile)
    except KeyError:
        raise arguments.ArgumentError(
            ("Unrecognized file type for data file {}").format(dfile)
        )

    dgroup = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, "data_group")

    conf_dict = None
    if fB_data.ftype == "ini":
        raise arguments.ArgumentError("ini type not supported for data")
    if fB_data.ftype == "csv":

        if fB_data.subkey is None:
            parse_str = ",F,Xr,Xi,S,E"
        else:
            parse_str = fB_data.subkey

        F_list = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, "frequency")
        Xr_list = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, "dataXreal")
        Xi_list = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, "dataXimaginary")
        SNR_list = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, "dataSNR")
        COH_list = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, "dataCOH")
        E_list = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, "dataEmphasis")
        parse_map = {
            "F": F_list[0],
            "Xr": Xr_list[0],
            "Xi": Xi_list[0],
            "S": SNR_list[0],
            "C": COH_list[0],
            "E": E_list[0],
            ".": None,
            "*": None,
            "?": None,
            "": None,
        }

        try:
            fdict_data = file_io.load_csv(
                fname=fB_data.fname,
                parse_str=parse_str,
                parse_map=parse_map,
            )
        except KeyError as e:
            k = e.args[0]
            raise arguments.ArgumentError(
                (
                    "CSV parse string '{}' has unrecognized key {}. Keys must be within"
                    " ['F', 'Xr', 'Xi', 'S', 'E', '.']"
                ).format(parse_str, k)
            )
        data_dict = fdict_data

        if mode == "printdata":
            dump_fdict_keys(data_dict, file=sys.stderr)
            raise SystemExit(0)
    elif fB_data.ftype == "special":
        fdict_data = file_io.load_any(
            fname=fB_data.fname,
            ftype=fB_data.ftype,
        )
        data_dict = fdict_data
        conf_dict = fdict_data
    else:
        fdict_data = file_io.load_any(
            fname=fB_data.fname,
            ftype=fB_data.ftype,
        )

        if mode == "printdata":
            print("data file layout for mode=printdata")
            dump_fdict_keys(fdict_data)
            raise SystemExit(0)

        dgroup_use = dgroup
        if fB_data.subkey is not None:
            dgroup_use = cmdline.mapcheck_cslist(fB_data.subkey)

        if len(dgroup_use) == 0:
            data_dict = fdict_data
            used_dgroups = [None]
        else:
            # detect and warn if more than one are found
            data_dict, used_dgroups = find_groups(fdict_data, dgroup_use)

        if len(used_dgroups) == 0:
            print(
                ("No data groups found in data file {}").format(fB_data.fname),
                file=sys.stderr,
            )
            print("The data file layout is:", file=sys.stderr)
            dump_fdict_keys(fdict_data, file=sys.stderr)
            # TODO dump data file layout
            raise SystemExit(1)
        elif len(used_dgroups) > 1:
            print(
                ("Warning: multiple possible data groups found in data file {}").format(
                    fB_data.fname
                ),
                file=sys.stderr,
            )
            print("Groups aggregated: {}".format(used_dgroups), file=sys.stderr)

        # detect and warn if more than one are found
        conf_dict, used_cgroups = find_groups(fdict_data, confB.cgroup)
        if len(used_cgroups) > 1:
            print(
                (
                    "Warning: multiple possible config groups found in data file {}"
                ).format(fB_data.fname),
                file=sys.stderr,
            )
            print("Groups aggregated: {}".format(used_cgroups), file=sys.stderr)

        if mode == "printconfig":
            print("data file layout for mode=printconfig, ignoring data groups")
            print("file: {}".fortmat(fB_data.fname))
            # pop out the used data groups. Only OK since we are exiting after
            # the print
            for dg in used_dgroups:
                fdict_data.pop(dg, None)
            dump_fdict_keys(fdict_data)
            raise SystemExit(0)

    ##debugging
    # dump_fdict_keys(data_dict, file = sys.stderr)

    # if any were loaded and conf_dict is not None:
    if conf_dict:
        kwfull.update(conf_dict)
        # args always takes precedent
        kwfull.update(args)

    return Bunch(
        fdict=fdict_data,
        data_dict=data_dict,
    )


def data_assemble(aid, kwfull, dataB):
    def pull_data_element(elname):
        """
        pulls a single element from a list of possible elements
        """
        clist = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, elname)
        val, used = find_elements(dataB.data_dict, clist, allow_dicts=False)
        if len(used) > 1:
            raise DataError(
                ("multiple possible redundant elements in data group: {}").format(used)
            )
        elif len(used) > 0:
            return val[0], used[0], clist
        else:
            return None, None, clist

    def pull_data_elements(elname):
        """
        pulls a multiple element from a list of possible elements
        """
        clist = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, elname)
        val, used = find_elements(dataB.data_dict, clist, allow_dicts=False)
        if len(used) > 0:
            return val, used, clist
        else:
            return None, None, clist

    xfer = None
    try:
        XF, XF_used, XF_list = pull_data_element("frequency")
        if XF is None:
            raise DataError(
                (
                    "Frequency must be specified as one of {} (set by --frequency)"
                ).format(XF_list)
            )

        Xc, Xc_used, Xc_list = pull_data_element("dataXcomplex")
        if Xc is not None:
            xfer = np.array(Xc)
            if not np.iscomplexobj(xfer):
                raise (
                    RuntimeError(
                        "Xfer found, but does not appear to be complex {}".format(xfer)
                    )
                )

        Xr, Xr_used, Xr_list = pull_data_element("dataXreal")
        Xi, Xi_used, Xi_list = pull_data_element("dataXimaginary")
        if Xr is not None and Xi is not None:
            if xfer is not None:
                raise DataError("redundant elements in data group")
            xfer = np.asarray(Xr) + 1j * np.asarray(Xi)
        elif Xr is not None:
            raise DataError("Data {} given but not imag part".format(Xr_used))
        elif Xi is not None:
            raise DataError("Data {} given but not real part".format(Xi_used))

        Xa, Xa_used, Xa_list = pull_data_element("dataXamplitude")
        Xdb, Xdb_used, Xdb_list = pull_data_element("dataXdB")
        if Xa is not None and Xdb is not None:
            raise DataError("redundant amplitude elements in data group")
        elif Xdb is not None:
            Xa = 10 ** (np.asarray(Xdb) / 20.0)

        Xp, Xp_used, Xp_list = pull_data_element("dataXphaseRad")
        Xd, Xd_used, Xd_list = pull_data_element("dataXphaseDeg")
        if Xp is not None and Xd is not None:
            raise DataError("redundant phase elements in data group")
        elif Xd is not None:
            Xp = np.asarray(Xd) * np.pi / 180

        if Xa is not None and Xp is not None:
            if xfer is not None:
                raise DataError("redundant elements in data group")
            xfer = np.asarray(Xa) * np.exp(1j * np.asarray(Xp))
        elif Xa is not None:
            raise DataError("Data amplitude given but not phase part")
        elif Xp is not None:
            raise DataError("Data phase given but not amplitude part")

        if xfer is None:
            raise DataError("No transfer function elements given in data group")

        XS, XS_used, XS_list = pull_data_elements("dataSNR")
        if XS is not None:
            if len(XS) > 1:
                XS = tuple(np.asarray(_) for _ in XS)
            else:
                XS = XS[0]
        XC, XC_used, XC_list = pull_data_elements("dataCOH")
        if XC is not None:
            if len(XC) > 1:
                XC = tuple(np.asarray(_) for _ in XC)
            else:
                XC = XC[0]
        XE, XE_used, XE_list = pull_data_elements("dataEmphasis")
        if XE is not None:
            if len(XE) > 1:
                XE = tuple(np.asarray(_) for _ in XE)
            else:
                XE = XE[0]
    except DataError as e:
        print("Error: {}".format(e.args[0]))
        print("Data group layout is: ")
        dump_fdict_keys(dataB.data_dict)
        print("Error: {}".format(e.args[0]))
        raise SystemExit(1)

    data_dict = dict(
        xfer=xfer,
        F_Hz=XF,
        emphasis=XE,
        SNR=XS,
        COH=XC,
    )
    return data_dict


def check_pzargs(args):
    p = args.get("poles", None)
    z = args.get("zeros", None)
    P = args.get("poles_overlay", None)
    Z = args.get("zeros_overlay", None)
    msg = (
        "Error, must specify some poles if -{0} flag given, "
        " could be parse ambiguity with '-', use -{0}=-#.##,-#.##... "
        " or -{0} -#.## -#.## -#.## "
    )
    if p is not None and len(p) == 0:
        print(msg.format("p"), file=sys.stderr)
        SystemExit(1)
    elif z is not None and len(z) == 0:
        print(msg.format("z"), file=sys.stderr)
        SystemExit(1)
    elif P is not None and len(P) == 0:
        print(msg.format("P"), file=sys.stderr)
        SystemExit(1)
    elif Z is not None and len(Z) == 0:
        print(msg.format("Z"), file=sys.stderr)
        SystemExit(1)


def IIRrationalV2fit(argv=None, **kwargs):
    """ """
    if argv is None:
        argv = sys.argv[1:]
    elif isinstance(argv, str):
        import shlex

        argv = shlex.split(argv)


    ap = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        add_help=False,
    )
    ap.add_argument('--cmdline_file', dest='cmdline_file', default=None)
    args, unknown_args = ap.parse_known_args(argv)
    if args.cmdline_file is not None:
        # print("COMMAND_LINE FILE: ", args.cmdline_file)
        # print(unknown_args)
        argv = unknown_args
        try:
            with open(args.cmdline_file) as F:
                cmdline_add = F.read()
                print("FILE: ", cmdline_add)
                import shlex
                argv2 = shlex.split(cmdline_add)
                print("argv2", argv2)
                argv = argv + argv2
        except FileNotFoundError:
            print("WARNING: file not found ", args.cmdline_file)
            pass

    aid = fit_aid.FitAid()

    kw_hints = dict(arguments.kw_hints)
    for name in arguments.standardargs.kw_hints_data.keys():
        del kw_hints[name]
    kw_hints.update(cmdline.kw_hints)

    ap = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
    )
    ap = pyargparse.kwdict_argparse(ap, kw_hints, groups_kw=cmdline.groups_kw)
    args = ap.parse_args(argv)
    args = vars(args)

    mode = arguments.grab_kwargs(aid, args, cmdline.kw_hints, "mode")
    kwfull = dict()

    confB = config_load(
        aid=aid,
        kwfull=kwfull,
        args=args,
        mode=mode,
    )

    ofile = args["outputfile"]
    if ofile == "":
        ofile = None
    overwrite = arguments.grab_kwargs(aid, args, cmdline.kw_hints, "overwrite")
    if ofile is not None:
        try:
            fB_output = file_io.determine_type(ofile)
        except KeyError:
            raise arguments.ArgumentError(
                ("Unrecognized file type for output file {}").format(ofile)
            )
        if not overwrite and os.path.exists(fB_output.fname):
            print(
                ("Error: output file {} already exists").format(ofile), file=sys.stderr
            )
            raise SystemExit(1)

    dataB = data_load(
        aid=aid,
        kwfull=kwfull,
        args=args,
        mode=mode,
        confB=confB,
    )

    if mode == "printsettings":
        print("settings dictionary after all loaded configurations")
        dump_fdict_keys(kwfull)
        raise SystemExit(0)

    # pull these early to make sure they are there and parsed before doing more work
    choose = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, "choose")
    information = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, "information")
    LIGO_foton = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, "LIGO_foton")
    refine = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, "refine")
    refine_file = arguments.grab_kwargs(aid, kwfull, cmdline.kw_hints, "refine_file")
    plot_fit = arguments.grab_kwargs(aid, args, cmdline.kw_hints, "plot_fit")
    plot_order = arguments.grab_kwargs(aid, args, cmdline.kw_hints, "plot_order")

    data_dict = data_assemble(
        aid=aid,
        kwfull=kwfull,
        dataB=dataB,
    )

    # Check these for error handing, since these can be zero for bad reasons
    check_pzargs(args=args)

    kwuse = dict(
        mode=mode,
    )
    arguments.transfer_kw(kwfull, kwuse, arguments.kw_hints)
    kwconfig = dict(kwuse)

    kwuse.update(data_dict)

    # TODO verbosity?
    print("Arguments: ")
    dump_fdict_keys(kwuse, depth=0)

    results = data2filter(
        # coding_map=fitters_ZPK.codings_s.coding_maps.SOS,
        # coding_map=fitters_ZPK.codings_s.coding_maps.SOSnl,
        **kwuse
    )

    version_dict = dict(
        data2filter="v2",
        version=IIRversion,
        # git_branch=IIRversion_auto.git_branch,
        # git_shorthash=IIRversion_auto.git_shorthash,
        # git_longhash=IIRversion_auto.git_longhash,
        # git_last_tag=IIRversion_auto.git_last_tag,
    )

    if plot_order is not None:
        try:
            print("plotting for -R, --plot_order={}".format(plot_order))
            results.investigate_order_plot(fname=plot_order)
        except Exception as E:
            print("Error: plotting exception {}".format(E), file=sys.stderr)

    if choose in ["shell", "prompt"]:
        choice_shell(
            results=results,
            data=DeepBunch(data_dict),
            config=DeepBunch(kwconfig),
            dfile=DeepBunch(dataB.fdict),
            cfiles=DeepBunch(confB.fdicts),
            version=DeepBunch(version_dict),
        )
    elif choose.startswith("auto"):
        remainder = choose[4:]
        if not remainder:
            order = 10
        else:
            try:
                order = int(remainder)
            except ValueError:
                print(
                    ("Error: auto order choice not an int, using auto10"),
                    file=sys.stderr,
                )
                order = 10
        order_arr = results.investigate_order_arrays().order
        num_lower = np.count_nonzero(order_arr < results.order)
        if num_lower > 0 and results.order > order:
            print(
                (
                    "Baseline order is {} > --choose=auto{}, and {} fit(s) are lower order, using shell"
                ).format(results.order, order, num_lower)
            )
            choice_shell(
                results=results,
                data=DeepBunch(data_dict),
                config=DeepBunch(kwconfig),
                dfile=DeepBunch(dataB.fdict),
                cfiles=DeepBunch(confB.fdicts),
                version=DeepBunch(version_dict),
            )
        else:
            print(
                (
                    "Baseline order is {} < --choose=auto{}, and {} fit(s) are lower order, using baseline"
                ).format(results.order, order, num_lower)
            )
    elif choose == 'best':
        results.choose(1000)
    elif choose.startswith("baseline"):
        remainder = choose[8:]
        if not remainder:
            order = None
        else:
            try:
                order = int(remainder)
            except ValueError:
                print(
                    ("Error: baseline order choice not an int, using baseline10"),
                    file=sys.stderr,
                )
                order = 10
        print("CHOOSING:", order)
        if order is not None and results.order > order:
            results.choose(order)
    else:
        try:
            order = int(choose)
        except ValueError:
            print(
                "Error: order choice not an int, using baseline order", file=sys.stderr
            )
        else:
            results.choose(order)

    if LIGO_foton in ["-", "None", "none"]:
        pass
    elif LIGO_foton == "Sf":
        print("-L, --LIGO_foton=Sf output:")
        print("%%%")
        print(results.as_foton_str_ZPKsf())
        print("%%%")
    else:
        print("Error: Unrecognized foton output type", file=sys.stderr)

    if refine:
        print(
            "%%% add or adjust the following poles and zeros to bootstrap the fit and refine. Disable this message with --no-refine"
        )
        print(results.as_refine_str())
        print("%%%")

    if refine_file:
        with open(refine_file, 'w') as F:
            F.write(results.as_refine_str())

    if plot_fit is not None:
        try:
            print("plotting for -r, --plot_fit={}".format(plot_fit))
            results.investigate_fit_plot(fname=plot_fit)
        except Exception as E:
            print("Error: plotting exception {}".format(E), file=sys.stderr)
            raise

    # TODO, need to ensure that this output form meets the common exchange
    # requirements for writing to the many filetypes
    output = collections.OrderedDict()
    output["fit"] = results.choice_export_dict()
    output["config"] = kwconfig
    output["fit_alternatives"] = results.alternatives_export_dict()
    output["data"] = data_dict
    output["version"] = version_dict
    # output['config_full']      = results.fit_aid.hints
    if information is not None:
        output["information"] = information

    if ofile is not None:
        file_io.write_any(
            fname=fB_output.fname,
            ftype=fB_output.ftype,
            fdict=output,
        )
    return results


def choice_shell(**kw):
    # First import the embed function
    from IPython.terminal.embed import InteractiveShellEmbed

    banner = padding_remove(
        """
    ------------------------------------------
    wield.iirrational fit choose investigation shell
    ------------------------------------------
    defined variables:
        results: result object of the wield.iirrational data2filter fit
        data:    loaded data dictionary
        config:  loaded config dictionary
        dfile:   original data file, normalized to dictionary form
        cfiles:  original config files, normalized to dictionary form
        exit:    function call exit() to leave shell

    use -j, --choose option to prevent dropping into this shell

    tab complete at ">results." to get methods of the results object
    to inspect the fit.

    typical usage:
    >results.investigate_order_plot()
    >results.choose(6)
    >exit()
    """
    )

    user_ns = dict(kw)

    ipshell = InteractiveShellEmbed(
        banner1=banner,
        user_ns=user_ns,
    )
    ipshell.enable_matplotlib()
    ipshell()


main = IIRrationalV2fit

if __name__ == "__main__":
    import sys

    IIRrationalV2fit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import re

_re_any_words = re.compile("\s*\S+")
_re_leading_space = re.compile("\s*")


def padding_remove(docstring_like, tabsize=4):
    """
    Both removes the leading line paddings on triple quoted strings and removes common indentation
    (expands tabs to do so)
    """
    docstring_like = docstring_like.strip()
    docstring_like = docstring_like.expandtabs(tabsize)
    docstring_lines = docstring_like.splitlines(True)

    spacings = []
    for line in docstring_lines[1:]:
        if len(line.strip()) == 0:
            continue
        spacings.append(len(_re_leading_space.match(line).group(0)))
    if spacings:
        common_spacing = min(spacings)
    else:
        common_spacing = 0
    common_tab_spacing = (common_spacing // tabsize) * tabsize

    if docstring_lines:
        fixed_lines = [docstring_lines[0]]
        for line in docstring_lines[1:]:
            fixed_lines.append(line[common_tab_spacing:])
    else:
        fixed_lines = []

    return "".join(fixed_lines)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np


def SNR_fix(aid):
    if aid.hint("trust_SNR"):
        return
    with aid.log_heading("SNR Fix Test"):
        W = aid.fitter.W
        SNR_max = aid.hint("SNR_max")
        if SNR_max is not None:
            Wclip = np.minimum(W, SNR_max)
            if np.any(Wclip < W):
                W = Wclip
                aid.fitter.W = Wclip
                aid.invalidate_fitters()
                aid.log_warn(
                    3,
                    """
                    Applying 'SNR_max'={} to clip the weights
                    """.format(
                        SNR_max
                    ),
                )

                # TODO, this validate=False is for relative_degree checks
                # since this is so early. Need a better mehtod
                aid.fitter_update(
                    representative=False,
                    validate=False,
                )

        effective_data_points = (np.sum(W ** 2)) ** 2 / np.sum(W ** 4)
        ratio = effective_data_points / len(W)
        Wmax = max(W)

        pref_ratio = aid.hint("SNR_regularize_ratio")
        pref_scale = aid.hint("SNR_regularize_scale")

        did_warn = False
        if ratio < 1 - pref_scale / Wmax:
            aid.log_warn(
                3,
                """
                The number of effective data points N=(ΣW^2)^2/(ΣW^4)={:.2e}*len(W)
                [where W=SNR] is below the configured 'SNR_regularize_scale'={},
                given the maximum SNR={}. Now Finding an SNR ceiling that balances
                the ratio with max SNR.
                """.format(
                    ratio, pref_scale, Wmax
                ),
            )
            did_warn = True
        elif ratio < pref_ratio:
            aid.log_warn(
                3,
                """
                The number of effective data points N=(ΣW^2)^2/(ΣW^4)={:.2e}*len(W)
                [where W=SNR] is below the configured 'SNR_regularize_ratio'={}.
                Now Finding an SNR ceiling to adjust to that ratio.
                """.format(
                    ratio, pref_ratio
                ),
            )
            did_warn = True

        if did_warn:
            aid.log_rationale(
                3,
                """
                Disable this adjustment by using a smaller value of
                'SNR_regularize_ratio', setting 'fix_SNR'=False, or
                setting 'trust_SNR'=True.
                """.format(
                    pref_ratio
                ),
            )
        else:
            return

        W = np.sort(aid.fitter.W)
        idx_min = 0
        idx_max = len(W)

        # binary search
        while idx_min + 1 < idx_max:
            idx_mid = (idx_max - idx_min) // 2 + idx_min
            W_ceiling = W[idx_mid]
            if W_ceiling <= 0:
                idx_min = idx_mid
                continue
            W_new = np.minimum(W_ceiling, W)
            effective_data_points = (np.sum(W_new ** 2)) ** 2 / np.sum(W_new ** 4)
            ratio = effective_data_points / len(W)
            pref_scale_ratio = 1 - pref_scale / W_ceiling
            if ratio < pref_ratio or ratio < pref_scale_ratio:
                idx_max = idx_mid
            else:
                idx_min = idx_mid
            aid.log_debug(
                9,
                "W:ratio",
                pref_ratio,
                pref_scale_ratio,
                ratio,
                idx_min,
                idx_max,
                W_ceiling,
            )

        W_ceiling = W[idx_min]
        aid.log_warn(
            3,
            """
            Using SNR<{} ceiling.
            """.format(
                W_ceiling
            ),
        )
        aid.fitter.W = np.minimum(aid.fitter.W, W_ceiling)
        aid.invalidate_fitters()

        # TODO, this validate=False is for relative_degree checks
        # since this is so early. Need a better mehtod
        aid.fitter_update(
            representative=False,
            validate=False,
        )


def check_sample_variance(aid):
    # TODO
    return


def check_sample_variance_adv(aid):
    # TODO
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import numpy as np
from wield.bunch import Bunch


def nearest_idx(
    lst_1,
    lst_2=None,
    metric_pair_dist=None,
    return_distances=False,
):
    """
    If lst_2 is given, this returns all of the nearest items in lst_2 to lst_1.
    If not given, this returns all of the nearest elements of lst_1 to itself,
    ignoring self elements.

    if metric_pair_dist is None, use the standard distance on complex plane.
    This is the fastest.
    """
    dists = []
    if lst_2 is not None:
        # TODO, this could be much more efficient with sorting..
        if metric_pair_dist is None:

            def metric_pair_dist(r1, r2):
                return abs(r1 - r2)

        nearest_lst = []
        for r1 in lst_1:
            if r1 is None:
                nearest_lst.append(None)
                continue
            dist_nearest = float("inf")
            idx_nearest = None
            for idx_2, r2 in enumerate(lst_2):
                if r2 is None:
                    continue
                dist = metric_pair_dist(r1, r2)
                if dist < dist_nearest:
                    idx_nearest = idx_2
                    dist_nearest = dist
            nearest_lst.append(idx_nearest)
            dists.append(dist_nearest)
    else:
        # TODO, this could be much more efficient with sorting..
        if metric_pair_dist is None:

            def metric_pair_dist(r1, r2):
                return abs(r1 - r2)

        nearest_lst = []
        for idx_1, r1 in enumerate(lst_1):
            if r1 is None:
                nearest_lst.append(None)
                continue
            dist_nearest = float("inf")
            idx_nearest = None
            for idx_2, r2 in enumerate(lst_1):
                if idx_2 == idx_1:
                    continue
                if r2 is None:
                    continue
                dist = metric_pair_dist(r1, r2)
                if dist < dist_nearest:
                    idx_nearest = idx_2
                    dist_nearest = dist
            nearest_lst.append(idx_nearest)
            dists.append(dist_nearest)
    if return_distances:
        return nearest_lst, dists
    else:
        return nearest_lst


def nearest_unique_idx(l1, l2):
    idx_p_used = []
    idx_list = []
    l1 = list(l1)

    for idx_1, idx_2 in enumerate(l1):
        if idx_2 is None:
            continue
        if idx_1 == l2[idx_2]:
            idx_p_used.append(idx_2)
            idx_list.append((idx_1, idx_2))
        else:
            l1[idx_1] = None
    l2_copy = [None] * len(l2)
    for idx_2 in idx_p_used:
        l2_copy[idx_2] = l2[idx_2]
    return Bunch(
        idx_list=idx_list,
        l1=l1,
        l2=l2_copy,
    )


def nearest_unique_pairs(
    l1,
    l2,
    metric_pair_dist=None,
):
    r12_list = []
    idx_list = []
    l1 = list(l1)
    l2 = list(l2)
    l1_nearest = nearest_idx(l1, l2, metric_pair_dist=metric_pair_dist)
    l2_nearest = nearest_idx(l2, l1, metric_pair_dist=metric_pair_dist)

    l1_remain = []
    l2_remain = []
    idx_2_used = []

    for idx_1, idx_2 in enumerate(l1_nearest):
        if idx_2 is None:
            l1_remain.append(l1[idx_1])
            continue
        # coding_z = aid.fitter.num_codings[idx_1]
        # coding_p = aid.fitter.den_codings[idx_2]
        # TODO annotate about stability
        p = l2[idx_2]
        z = l1[idx_1]
        if idx_1 == l2_nearest[idx_2]:
            idx_2_used.append(idx_2)
            r12_list.append((z, p))
            idx_list.append((idx_1, idx_2))
        else:
            l1_remain.append(l1[idx_1])
            l1_nearest[idx_1] = None
    idx_2_used = set(idx_2_used)
    for idx_2, p in enumerate(l2):
        if idx_2 not in idx_2_used:
            l2_remain.append(p)
            l2_nearest[idx_2] = None
    assert len(r12_list) + len(l1_remain) == len(l1)
    assert len(r12_list) + len(l2_remain) == len(l2)
    return Bunch(
        r12_list=r12_list,
        l1_remain=l1_remain,
        l2_remain=l2_remain,
        idx_list=idx_list,
        l1=l1_nearest,
        l2=l2_nearest,
    )


def nearest_pairs(
    l1,
    l2,
    metric_pair_dist=None,
):
    # TODO, allow other rankings than distance

    rpB = nearest_unique_pairs(l1, l2, metric_pair_dist)
    # not going to maintain these lists
    del rpB.idx_list
    del rpB.l1
    del rpB.l2

    while True:
        pair_lists = []
        l1_nearest, l1_dist = nearest_idx(
            rpB.l1_remain,
            rpB.l2_remain,
            metric_pair_dist=metric_pair_dist,
            return_distances=True,
        )
        for idx_1, idx_2 in enumerate(l1_nearest):
            if idx_2 is None:
                continue
            dist = l1_dist[idx_1]
            pair_lists.append((dist, idx_1, idx_2))
        l2_nearest, l2_dist = nearest_idx(
            rpB.l2_remain,
            rpB.l1_remain,
            metric_pair_dist=metric_pair_dist,
            return_distances=True,
        )
        for idx_2, idx_1 in enumerate(l2_nearest):
            if idx_1 is None:
                continue
            dist = l2_dist[idx_2]
            pair_lists.append((dist, idx_1, idx_2))
        if not pair_lists:
            break
        pair_lists.sort()
        dist, idx_1, idx_2 = pair_lists[0]
        rpB.r12_list.append((rpB.l1_remain[idx_1], rpB.l2_remain[idx_2]))
        del rpB.l1_remain[idx_1]
        del rpB.l2_remain[idx_2]
    return rpB


def SOS_pair_rolloff(
    Zr,
    Zc,
    Pr,
    Pc,
    metric_rolloff=abs,
):
    metric = metric_rolloff
    zzpp_list = []

    Zr = list(Zr)
    Zc = list(Zc)
    Pr = list(Pr)
    Pc = list(Pc)

    # sort on their rolloff bandwidth
    Zc.sort(key=metric)
    Zr.sort(key=metric)
    Pc.sort(key=metric)
    Pr.sort(key=metric)

    def real_low(Rr, Rc):
        if Rc:
            if Rr:
                if metric(Rr[0]) < metric(Rc[0]):
                    r_low = True
                else:
                    r_low = False
            else:
                r_low = False
        else:
            r_low = True
        return r_low

    def real_hold_insert(rpair, rhold):
        if rhold is None:
            return rpair
        z1, p1 = rhold
        z2, p2 = rpair
        zzpp_list.append((z1, z2, p1, p2))
        return None

    rpair_hold = None
    while True:
        Pr_low = real_low(Pr, Pc)
        Zr_low = real_low(Zr, Zc)

        if Pr_low and Zr_low:
            if Pr and Zr:
                p = Pr.pop(0)
                z = Zr.pop(0)
                rpair_hold = real_hold_insert((z, p), rpair_hold)
                continue
            elif Pr:
                # can only occur if Zc is also empty
                p = Pr.pop(0)
                rpair_hold = real_hold_insert((None, p), rpair_hold)
                continue
            elif Zr:
                # can only occur if Pc is also empty
                z = Zr.pop(0)
                rpair_hold = real_hold_insert((z, None), rpair_hold)
                continue
            else:
                # can only occur when all lists are empty
                break

        elif not Pr_low and not Zr_low:
            if Pc and Zc:
                p = Pc.pop(0)
                z = Zc.pop(0)
                zzpp_list.append((z, z.conjugate(), p, p.conjugate()))
                continue
            elif Pc:
                # can only occur if Zr is also empty
                p = Pc.pop(0)
                zzpp_list.append((None, None, p, p.conjugate()))
                continue
            elif Zc:
                # can only occur if Pr is also empty
                z = Zc.pop(0)
                zzpp_list.append((z, z.conjugate(), None, None))
                continue
            else:
                # can only occur when all lists are empty
                break

        # now it must be a mix for Pr_low and Zr_low, so work around the complex one
        elif Pr_low:
            # must be that Zr_low is False, Pr_low is true
            z = Zc.pop(0)
            if Pc:
                pc_t = Pc[0]
                if len(Pr) > 1:
                    pr_t = Pr[1]
                    if metric(pr_t) < metric(pc_t):
                        # we have a DOUBLE low real, so use that
                        p1 = Pr.pop(0)
                        p2 = Pr.pop(0)
                        zzpp_list.append((z, z.conjugate(), p1, p2))
                        continue
                    else:
                        # complex is between the others, so use it instead
                        p = Pc.pop(0)
                        zzpp_list.append((z, z.conjugate(), p, p.conjugate()))
                        continue
                else:
                    # if Pr is almost empty, then favor the complex
                    p = Pc.pop(0)
                    zzpp_list.append((z, z.conjugate(), p, p.conjugate()))
                    continue
            else:
                # no more complex poles, so must exhaust real ones
                if len(Pr) > 1:
                    p1 = Pr.pop(0)
                    p2 = Pr.pop(0)
                    zzpp_list.append((z, z.conjugate(), p1, p2))
                    continue
                elif len(Pr) == 1:
                    p1 = Pr.pop(0)
                    zzpp_list.append((z, z.conjugate(), p1, None))
                    continue
                else:
                    zzpp_list.append((z, z.conjugate(), None, None))
                    continue
        else:
            # must be that Pr_low is False
            # must be that Zr_low is False, Pr_low is true
            p = Pc.pop(0)
            if Zc:
                zc_t = Zc[0]
                if len(Zr) > 1:
                    zr_t = Zr[1]
                    if metric(zr_t) < metric(zc_t):
                        # we have a DOUBLE low real, so use that
                        z1 = Zr.pop(0)
                        z2 = Zr.pop(0)
                        zzpp_list.append((z1, z2, p, p.conjugate()))
                        continue
                    else:
                        # complex is between the others, so use it instead
                        z = Zc.pop(0)
                        zzpp_list.append((z, z.conjugate(), p, p.conjugate()))
                        continue
                else:
                    # if Zr is almost empty, then favor the complex
                    z = Zc.pop(0)
                    zzpp_list.append((z, z.conjugate(), p, p.conjugate()))
                    continue
            else:
                # no more complex zeros, so must exhaust real ones
                if len(Zr) > 1:
                    z1 = Zr.pop(0)
                    z2 = Zr.pop(0)
                    zzpp_list.append((z1, z2, p, p.conjugate()))
                    continue
                elif len(Zr) == 1:
                    z1 = Zr.pop(0)
                    zzpp_list.append((z1, None, p, p.conjugate()))
                    continue
                else:
                    zzpp_list.append((None, None, p, p.conjugate()))
                    continue

    if rpair_hold is not None:
        z, p = rpair_hold
        zzpp_list.append((z, None, p, None))

    return zzpp_list


def match_SOS_pairs(
    Zr,
    Zc,
    Pr,
    Pc,
    F_nyquist_Hz=None,
    metric_rolloff=None,
    metric_pair_dist=None,
):
    """
    Match and create pairs suitable for SOS representation. The output
    is a list of 4-tuples with z1, z2, p1, p2. If roots are complex,
    they are guaranteed to be partnered with their conjugate.


    Some z, or p may be None, indicating that the system ran out.
    """

    Nz_tot = len(Zc) * 2 + len(Zr)
    Np_tot = len(Pc) * 2 + len(Pr)

    if metric_pair_dist is None:
        if F_nyquist_Hz is None:

            def metric_pair_dist(r1, r2):
                return abs(r1 - r2)

        else:

            def metric_pair_dist(r1, r2):
                return abs(r1 - r2)

    pairB = nearest_unique_pairs(Zc, Pc, metric_pair_dist=metric_pair_dist)
    Zc = pairB.l1_remain
    Pc = pairB.l2_remain

    zzpp_list = []
    for z, p in pairB.r12_list:
        zzpp_list.append((z, z.conjugate(), p, p.conjugate()))

    if metric_rolloff is None:
        if Np_tot < Nz_tot:
            # when the filter is AC coupled, apply strongest
            # AC coupling early and weaker later
            if F_nyquist_Hz is None:

                def metric_rolloff(r):
                    return abs(r)

            else:

                def metric_rolloff(r):
                    F = np.angle(r)
                    d = 1 - abs(r)
                    return (d ** 2 + F ** 2) ** 0.5

        else:
            # when the filter has rolloff, apply the strongest rolloff later
            if F_nyquist_Hz is None:

                def metric_rolloff(r):
                    return 1 / abs(r)

            else:

                def metric_rolloff(r):
                    F = np.angle(r)
                    d = 1 - abs(r)
                    return 1 / (d ** 2 + F ** 2) ** 0.5

    zzpp_list.extend(SOS_pair_rolloff(Zr, Zc, Pr, Pc, metric_rolloff=metric_rolloff))

    Nz = 0
    Np = 0
    for (z1, z2, p1, p2) in zzpp_list:
        if z1 is not None:
            Nz += 1
        if z2 is not None:
            Nz += 1
        if p1 is not None:
            Np += 1
        if p2 is not None:
            Np += 1
    assert Nz == Nz_tot
    assert Np == Np_tot

    def sorter(zzpp):
        (z1, z2, p1, p2) = zzpp
        lst = []
        for r in zzpp:
            if r is not None:
                lst.append(abs(r.imag))
        if lst:
            return np.average(lst)
        return 0
    zzpp_list.sort(key=sorter)
    return zzpp_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import collections
import sys
import re
import yaml


def yaml_load(fname):
    # HACK: fix loading number in scientific notation
    #
    # https://stackoverflow.com/questions/30458977/yaml-loads-5e-6-as-string-and-not-a-number
    #
    # An apparent bug in python-yaml prevents it from regognizing
    # scientific notation as a float.  The following is a modified version
    # of the parser that recognize scientific notation appropriately.
    yaml_loader = yaml.SafeLoader
    yaml_loader.add_implicit_resolver(
        "tag:yaml.org,2002:float",
        re.compile(
            """^(?:
        [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
        |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
        |\\.[0-9_]+(?:[eE][-+][0-9]+)?
        |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
        |[-+]?\\.(?:inf|Inf|INF)
        |\\.(?:nan|NaN|NAN))$""",
            re.X,
        ),
        list("-+0123456789."),
    )

    with open(fname, "r") as F:
        fdict = yaml.load(F, Loader=yaml_loader)
    return fdict


def dict_representer(dumper, data):
    return dumper.represent_dict(data.iteritems())


yaml.SafeDumper.add_representer(collections.OrderedDict, dict_representer)


def yaml_write(fname, fdict):
    if sys.version < (3, 4):
        with open(fname, "w") as F:
            yaml.safe_dump(fdict, F)
    else:
        with open(fname, "w", encoding="utf8") as F:
            yaml.safe_dump(fdict, F)



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

import numpy as np
from wield.iirrational.representations.polynomials import chebychev

from numpy.polynomial import chebyshev

chebroots = chebyshev.chebroots
chebfromroots = chebyshev.chebfromroots
chebder = chebyshev.chebder
chebadd = chebyshev.chebadd
chebmul = chebyshev.chebmul
chebdiv = chebyshev.chebdiv
chebval = chebyshev.chebval


def abs_sq(v):
    return np.real(v) ** 2 + np.imag(v) ** 2


test_roots = [
    0.5
    * np.asarray(
        (
            -2,
            -1.1 + 1j,
            -1.1 + 1j,
            -1.1 + 1j,
            -2 + 1j,
            -1 + 1j,
            -1 + 1j,
            -1 + 1j,
            -1.1 - 1j,
            -1.1 - 1j,
            -1.1 - 1j,
            -2 + 1j,
            -1 - 1j,
            -1 - 1j,
            -1 - 1j,
        )
    ),
]


@pytest.mark.parametrize("roots", test_roots)
def test_root_chaser(roots):
    c = chebfromroots(roots)
    c_orig = c

    def calc(z, pr=False):
        cval = chebval(z, c)
        cdval = chebval(z, cd)
        rat1 = cval / cdval
        ret = None
        if np.any(abs_sq(cdval) > tol):
            ret = rat1
        cd2val = chebval(z, cd2)
        rat2 = cdval / cd2val
        if ret is None and np.any(abs_sq(cd2val) > tol):
            ret = rat2
        cd3val = chebval(z, cd3)
        rat3 = cd2val / cd3val
        if ret is None and np.any(abs_sq(cd3val) > tol):
            ret = rat3
        cd4val = chebval(z, cd4)
        rat4 = cd3val / cd4val
        if ret is None:
            ret = rat4
        ret = rat2 * 1
        if pr:
            # print(abs(cval), abs(cdval), abs(cd2val), abs(cd3val))
            print(abs(rat1), abs(rat2), abs(rat3), abs(rat4))
        return ret

    def laguerre(z):
        cval = chebval(z, c)
        cdval = chebval(z, cd)
        cd2val = chebval(z, cd2)

        G = cdval / cval
        Gsq = G ** 2
        H = Gsq - cd2val / cval
        k = 2
        m = 2
        n = m + k

        disc = k / m * (n * H - Gsq)
        disc_root = disc ** 0.5
        a_p = n / (G + disc_root)
        a_n = n / (G - disc_root)
        if abs(a_p) < abs(a_n):
            return a_p
        else:
            return a_n

    print()

    tol = 1e-8
    current_z = (-1.01 + 1.01j) / 2
    roots = chebroots(c)
    select = np.nonzero(abs(roots - current_z) < 0.05)[0]
    print(select[1:])
    idx_update = 0

    for n in range(1):
        for idx_update in range(len(select)):
            myroot_idx = select[idx_update]
            subroots = np.concatenate(
                [roots[select[:idx_update]], roots[select[idx_update + 1 :]]]
            )
            print(subroots)
            # subroots = [
            #    -1 + 1j,
            #    -1 + 1j,
            # ]
            c = c_orig
            # for sr in subroots:
            #    sc = chebfromroots([sr])
            #    q, r = chebdiv(c, sc)
            #    print('remainder', abs(r), r)
            #    c = q
            # rescale by the root density nearest to the origin
            c = c / c[1]
            cd = chebder(c)
            cd2 = chebder(cd)
            cd3 = chebder(cd2)
            cd4 = chebder(cd3)

            current_z = (-1.01 + 1.01j) / 2
            current_z = -1.1
            scoots = []
            for idx in range(100):
                rat = calc(current_z, pr=True)
                # print("CVAL: ", abs(rat))
                # diff_z = -scale * rat
                # new_z = current_z + diff_z
                # val = calc(new_z)
                # a, b = np.polyfit(diff_z, val, deg = 1)
                # scoot = -b/a
                # scoot = scoot * abs(rat) / (abs(scoot)/100 + abs(rat))
                # scoot = -rat
                scoot = -laguerre(current_z)
                # scoot = -rat
                scoots.append(scoot)
                # print('scoot', scoot)
                current_z = current_z + scoot
            scoots = np.asarray(scoots)
            print(abs(scoots))
            roots[myroot_idx] = current_z
            print(myroot_idx, current_z)
            # print(abs(scoots))

    # assert(False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import numpy as np
import warnings
from wield.bunch.depbunch import depB_property, NOARG
from wield.bunch import Bunch
import collections

from .. import TFmath

from .rational_bases import DataFilterBase


lnGtup = collections.namedtuple("lnGtup", ["sign", "lnG"])
lnGxfer = collections.namedtuple("lnGxfer", ["xfer", "lnG"])
lnGVand = collections.namedtuple("lnGVand", ["vander", "lnG"])


class PolyFit(DataFilterBase):
    phase_missing = False
    # must fill
    poly = None
    roots_split_strict = False
    stabilize_method = "flip"

    def __build__(
        self,
        _args=None,
        npoles=NOARG,
        nzeros=NOARG,
        mindelay=False,
        stabilize=False,
        parent=NOARG,
        **kwargs
    ):
        if _args is not None:
            raise RuntimeError("Only keyword arguments allowed")
        if parent is not NOARG:
            if npoles is NOARG:
                npoles = parent.npoles
            if nzeros is NOARG:
                nzeros = parent.nzeros
            if mindelay is NOARG:
                npoles = parent.mindelay
            if nzeros is NOARG:
                stabilize = parent.stabilize

        self.npoles = npoles
        self.nzeros = nzeros
        self.mindelay = mindelay
        self.stabilize = stabilize

        super(PolyFit, self).__build__(parent=parent, **kwargs)
        # use these for now
        # self.lnG = 0, 1
        # self.gain = 1

        ##first to prevent lookups that shouldn't happen
        # self.zeros     = []
        # self.poles     = []
        # print(self.zeros)
        # print(self.poles)
        # self.a_vec
        # self.b_vec

    @depB_property
    def lnG(self, val=NOARG):
        if val is NOARG:
            raise RuntimeError("shouldn't compute")
            return self.b_vec[-1] / self.a_vec[-1]
        else:
            sgn, lnG = val
            if not np.isfinite(lnG):
                raise RuntimeError("Gain must be finite")
            return lnGtup(sgn, lnG)

    @depB_property
    def gain(self, val=NOARG):
        if val is NOARG:
            return self.lnG.sign * np.exp(self.lnG.lnG)
        else:
            if val == 0:
                raise RuntimeError("Gain 0 not allowed")
            if val.imag != 0:
                raise RuntimeError("Only Real Gain allowed")
            elif not np.isfinite(val):
                raise RuntimeError("Gain Must be finite")
            elif val > 0:
                self.lnG = 1, np.log(val)
            else:
                self.lnG = -1, np.log(-val)
            self.dependencies("lnG")
            return val

    @depB_property
    def a_vec(self, pvec=NOARG):
        if pvec is NOARG:
            vec = self.poly_fromroots(self.poles)
            norm = self.phi_coeff_norm(vec)
            if norm != 1:
                warnings.warn("poly normalization unexpected, {0}".format(norm))
            return vec
        else:
            pvec = np.asarray(pvec)
            # adjust gain numbers and canonicalize
            signG, lnG = self.get_raw("lnG")
            cG = self.phi_coeff_norm(pvec)
            self.lnG = signG * np.sign(cG), lnG - np.log(abs(cG))
            pvec = pvec / cG

            if self.get_raw("stabilize"):
                poles = self.poly.roots_rB(pvec, self.poly_constraints)
                # only prevent using this vector if needed, stability is helped if we don't go convert between roots and vectors too much
                # actually set it so that we get the stabilize action
                self.poles = poles
                # don't store any value as it should be generated from the poles if needed
                return self.TAG_NO_SET

            self.dependencies("poles")
            return pvec

    @depB_property
    def b_vec(self, pvec=NOARG):
        if pvec is NOARG:
            vec = self.poly_fromroots(self.zeros)
            norm = self.phi_coeff_norm(vec)
            if norm != 1:
                warnings.warn("poly normalization unexpected, {0}".format(norm))
            return vec
        else:
            pvec = np.asarray(pvec)

            signG, lnG = self.get_raw("lnG")
            cG = self.phi_coeff_norm(pvec)
            self.lnG = signG * np.sign(cG), lnG + np.log(abs(cG))
            pvec = pvec / cG

            if self.get_raw("mindelay"):
                zeros = self.poly.roots_rB(pvec, self.poly_constraints)
                # only prevent using this vector if needed, stability is helped if we don't go convert between roots and vectors too much
                # actually set it so that we get the stabilize action
                self.zeros = zeros
                return self.TAG_NO_SET
            # since they had to be computed anyway for the gain adjustment
            self.dependencies("zeros")
            return pvec

    @depB_property
    def zeros(self, val=NOARG):
        if val is NOARG:
            rB = self.poly.roots_rB(self.b_vec, self.poly_constraints)
            return self.RBalgo.expect(rB, self.root_constraint)
        else:
            # get previous
            rB = self.RBalgo.expect(val, self.root_constraint)
            if self.get_raw("mindelay"):
                rB = self.root_stabilize(
                    rB,
                    method=self.stabilize_method,
                )
            self.dependencies("b_vec")
            return rB

    @depB_property
    def poles(self, val=NOARG):
        if val is NOARG:
            rB = self.poly.roots_rB(self.a_vec, self.poly_constraints)
            return self.RBalgo.expect(rB, self.root_constraint)
        else:
            rB = self.RBalgo.expect(val, self.root_constraint)
            if self.get_raw("stabilize"):
                rB = self.root_stabilize(
                    rB,
                    method=self.stabilize_method,
                )
            self.dependencies("a_vec")
            return rB

    @depB_property
    def zeros_overlay(self, val=NOARG):
        if val is NOARG:
            return self.RBalgo.expect([], self.root_constraint)
        else:
            rB = self.RBalgo.expect(val, self.root_constraint)
            return rB

    @depB_property
    def poles_overlay(self, val=NOARG):
        if val is NOARG:
            return self.RBalgo.expect([], self.root_constraint)
        else:
            rB = self.RBalgo.expect(val, self.root_constraint)
            return rB

    @depB_property
    def V_b(self):
        return lnGVand(*self.poly.vander_lnG(self.X_grid, self.nzeros))

    @depB_property
    def V_a(self):
        return lnGVand(*self.poly.vander_lnG(self.X_grid, self.npoles))

    @depB_property
    def h_delay(self):
        return np.exp(-2j * np.pi * self.F_Hz * self.delay_s)

    @depB_property
    def h_a(self):
        self.dependencies("poles", "a_vec")
        X = self.X_grid
        pval = self.get_default("poles", None)
        # pval = self.poles
        if pval is not None:
            val, lnG = pval.val_lnG(X)
            adj = self.phi_gain_adj_roots(pval)
            # print('Ap', np.max(abs(val)), lnG)
            return lnGxfer(adj * val * self._phasing_roots(pval), lnG)

        a_vec = self.get_default("a_vec", None)
        if a_vec is not None:
            X = self.X_grid
            val, lnG = self.poly.val_lnG(X, a_vec)
            adj = self.phi_gain_adj_pvec(a_vec)
            # print('Bp', np.max(abs(val)), lnG)
            return lnGxfer(adj * val * self._phasing_vec(a_vec), lnG)

        # otherwise compute it from the preferred poles representation
        pval = self.poles
        val, lnG = pval.val_lnG(X)
        adj = self.phi_gain_adj_roots(pval)
        # print('Cp', np.max(abs(val)), lnG)
        return lnGxfer(adj * val * self._phasing_roots(pval), lnG)

    @depB_property
    def h_b(self):
        self.dependencies("zeros", "b_vec", "lnG")
        X = self.X_grid
        zval = self.get_default("zeros", None)
        # zval = self.zeros
        if zval is not None:
            val, lnG = zval.val_lnG(X)
            adj = self.phi_gain_adj_roots(zval)
            # print('Az', np.max(abs(val)), lnG)
            return lnGxfer(adj * val * self._phasing_roots(zval), lnG)

        b_vec = self.get_default("b_vec", None)
        if b_vec is not None:
            val, lnG = self.poly.val_lnG(X, b_vec)
            adj = self.phi_gain_adj_pvec(b_vec)
            # print('Bz', np.max(abs(val)), lnG)
            return lnGxfer(adj * val * self._phasing_vec(b_vec), lnG)

        # otherwise compute it from the preferred poles representation
        zval = self.zeros
        val, lnG = zval.val_lnG(X)
        adj = self.phi_gain_adj_roots(zval)
        # print('Cz', np.max(abs(val)), lnG)
        return lnGxfer(adj * val * self._phasing_roots(zval), lnG)

    @depB_property
    def _extra_xfer_fit(self):
        if self.F_nyquist_Hz is None:
            xfer_p, lnG_p = self.poles_overlay.val_lnG(self.X_grid)
            xfer, lnG = self.zeros_overlay.val_lnG(
                self.X_grid, h=1 / xfer_p, lnG=-lnG_p
            )
        else:
            xfer_p, lnG_p = self.poles_overlay.val_lnG(self.Xn_grid)
            xfer, lnG = self.zeros_overlay.val_lnG(
                self.Xn_grid, h=1 / xfer_p, lnG=-lnG_p
            )
        return xfer * np.exp(lnG)

    @depB_property
    def data_effective(self):
        return self.data / (self._extra_xfer_fit * self.h_delay)

    @depB_property
    def xfer_fit(self):
        return self._extra_xfer_fit * (
            self.lnG.sign
            * np.exp(self.lnG.lnG - self.h_a.lnG + self.h_b.lnG)
            * (self.h_b.xfer * self.h_delay)
            / self.h_a.xfer
        )

    @depB_property
    def residuals(self):
        debias_reweight = 1 / (0.001 + self.W ** 2)
        retB = Bunch()
        R = self.xfer_fit / self.data
        retB.resP = self.W * (R - 1)
        retB.resZ = self.W * (1 / R - 1)
        retB.resD = self.W * (R - 1 / R)
        retB.resD_average = np.sum(TFmath.abs_sq(retB.resD)) / (2 * len(self.data))
        retB.average = np.sum(
            (abs(retB.resP) ** 2 + abs(retB.resZ * debias_reweight) ** 2)
            / (1 + debias_reweight)
        ) / (2 * len(self.data))
        return retB

    @depB_property
    def residuals_average(self):
        return self.residuals.resD_average

    @depB_property
    def A_z(self):
        return self.V_b.vander * (
            (self._phasing_bfit() * self.lnG.sign)
            * np.exp(self.lnG.lnG - self.h_a.lnG + self.V_b.lnG)
            * self.W
            / (self.data_effective)
            / self.h_a.xfer
        ).reshape(-1, 1)

    @depB_property
    def A_zp(self):
        return self.V_a.vander * (
            (self._phasing_bfit() * self.lnG.sign)
            * np.exp(self.lnG.lnG - 2 * self.h_a.lnG + self.V_a.lnG + self.h_b.lnG)
            * self.W
            / (self.data_effective)
            * self.h_b.xfer
            / self.h_a.xfer ** 2
        ).reshape(-1, 1)

    @depB_property
    def A_p(self):
        return self.V_a.vander * (
            (self._phasing_afit() * self.lnG.sign)
            * np.exp(-self.lnG.lnG - self.h_b.lnG + self.V_a.lnG)
            * self.W
            * (self.data_effective)
            / self.h_b.xfer
        ).reshape(-1, 1)

    @depB_property
    def A_pz(self):
        return self.V_b.vander * (
            (self._phasing_afit() * self.lnG.sign)
            * np.exp(-self.lnG.lnG - 2 * self.h_b.lnG + self.h_a.lnG + self.V_b.lnG)
            * self.W
            * (self.data_effective)
            * self.h_a.xfer
            / self.h_b.xfer ** 2
        ).reshape(-1, 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import numpy as np
from wield.bunch import Bunch

from ..utilities import ensure_aid

from ..utilities import np as np_utils
from .. import representations
from .. import fitters_ZPK

from . import algorithms


def fitter_log_mag_generate(aid):
    # TODO
    return


def fitter_stabilized_log_mag_generate(aid):
    # TODO
    return


def root_stabilize(aid):
    z_c = np.copy(aid.fitter.zeros.c)
    z_r = np.copy(aid.fitter.zeros.r)
    z_c.real[z_c.real > 0] *= -1
    z_r.real[z_r.real > 0] *= -1

    p_c = np.copy(aid.fitter.poles.c)
    p_r = np.copy(aid.fitter.poles.r)
    p_c.real[p_c.real > 0] *= -1
    p_r.real[p_r.real > 0] *= -1

    aid.fitter_update(
        aid.fitter.regenerate(
            ZPKrep=aid.fitter.ZPKrep,
            coding_map=fitters_ZPK.codings_s.coding_maps.SOSsafe,
            z_c=z_c,
            z_r=z_r,
            p_c=p_c,
            p_r=p_r,
        ),
        representative=False,
    )
    return


def phase_patch(aid):
    fitter_mag = aid.fitter.copy()
    fitter_mag.residuals_log_im_scale = 0
    with fitter_mag.with_codings_only([fitter_mag.gain_coding]):
        fitter_mag.optimize_NM(residuals_type="log")
    fitter_mag.optimize(
        aid=aid,
        residuals_type="log",
    )
    res_avg = fitter_mag.residuals_average
    aid.log_debug(5, "log residuals average", res_avg)

    Wsort = np.sort(aid.fitter.W)
    Wthresh = Wsort[int(len(aid.fitter.W) * 0.2)]
    # TODO decide about this
    Wselect = aid.fitter.W >= 0 * Wthresh / 2
    data = fitter_mag.data / fitter_mag.xfer_fit
    # data = data / abs(data)
    xres = representations.ZPKwData(
        data=data[Wselect],
        W=aid.fitter.W[Wselect],
        F_Hz=aid.fitter.F_Hz[Wselect],
        F_nyquist_Hz=None,
    )

    fitter_X = fitters_ZPK.ZPKrep2MRF(
        ZPKrep=xres,
        delay_s_min=0,
        delay_s_max=0,
    )
    fitter_X.gain = 1
    # only use the first fraction of the data, the latter data is the most influenced
    # by delay and by bad fitting
    algorithms.sign_check_flip(fitter_X, max_N=len(fitter_X.F_Hz) // 3 + 3)
    fitter_X.optimize(residuals_type="zeros")
    fitter_X.delay_coding = fitters_ZPK.codings_cmn.CodingDelay(fitter_X)
    fitter_X.optimize(residuals_type="log")

    if aid.hint("trust_SNR"):
        weight_scale = 1
    else:
        with aid.log_heading("sample variance (from magnitude)"):
            R = fitter_X.data / fitter_X.xfer_fit
            W = fitter_X.W
            meanWeight = (np.sum(W ** 4) / np.sum(W ** 2)) ** 0.5
            cRvar = (np.sum((abs(R) - 1) ** 2 * W ** 2) / np.sum(W ** 2)) ** 0.5
            aid.log_debug(5, "variance", 1 / cRvar, meanWeight)
            weight_scale_true = meanWeight * cRvar
            weight_scale = weight_scale_true ** 0.5
            aid.log_alert(3, "Weight Scaling determined: ", weight_scale_true)
            aid.log_alert(3, "Weight Scaling Used ", weight_scale)

    phase_fixes = []
    with aid.log_heading("Phase patching"):
        shift_saves = []
        separations = list(
            np_utils.logspaced(
                3, len(fitter_X.F_Hz) / 2, int(1 + np.log(len(fitter_X.F_Hz)))
            )
        )
        check_second_pass = False
        while separations:
            sep = separations.pop()
            sep = int(sep)

            f = fitter_X.F_Hz
            f_d = (f[2:] - f[:-2]) / 2
            W = fitter_X.W[1:-1]
            R = (fitter_X.data / fitter_X.xfer_fit)[1:-1]
            f = f[1:-1]

            cR = np.cumsum(R / abs(R) * W ** 2 * f_d)
            cW = np.cumsum(W ** 2 * f_d)
            cFd = np.cumsum(f_d)

            shift_save = Bunch()
            # distance_10
            shift = (cR[sep:] - cR[0:-sep]) / (cW[sep:] - cW[0:-sep])
            shift_err = (
                weight_scale
                * ((cFd[sep:] - cFd[0:-sep]) / (cW[sep:] - cW[0:-sep]) / (2 * sep))
                ** 0.5
            )
            f_local = (f[sep:] + f[:-sep]) / 2
            bw_count = 0
            for sec_L, sec_R in np_utils.generate_sections(shift + shift_err < 0):
                loc_Hz = (f_local[sec_R] + f_local[sec_L]) / 2
                BW_Hz = (f_local[sec_R] - f_local[sec_L]) / 2
                aid.log(6, "LOCATION ", loc_Hz)
                aid.log(6, "BW: ", BW_Hz)

                try:
                    select2BW = (f_local < loc_Hz + 2 * BW_Hz) & (
                        f_local > loc_Hz - 2 * BW_Hz
                    )
                    Imax = np.max(shift.imag[select2BW])
                    Imin = np.min(shift.imag[select2BW])
                    Isep = Imax - Imin
                    Icmn = (Imax + Imin) / 2
                    aid.log(8, "Isep", Isep)
                    aid.log(8, "Icmn", Icmn)

                    if abs(Isep) + shift_err[sec_R] + shift_err[sec_L] < 0.75:
                        aid.log_debug(8, "Skipping Pair")
                        continue
                except ValueError:
                    aid.log_debug(8, "Skipping Pair, phase patch missing")
                    continue

                bw_count += 1
                if check_second_pass:
                    continue

                coding = fitters_ZPK.codings_cmn.CodingDelayPair(fitter_X)
                coding.F_Hz = loc_Hz
                coding.BW_Hz = BW_Hz
                fitter_X.num_codings.append(coding)
                fitter_X.codings_revision += 1

            if not check_second_pass:
                shift_save.sep = sep
                shift_save.f_local = f_local
                shift_save.shift = shift
                shift_save.shift_err = shift_err
                shift_saves.append(shift_save)
                fitter_X.optimize()
                # switch to second pass mode check
                check_second_pass = True
                # add the separation back for the second pass
                separations.append(sep)
            else:
                if bw_count > 0:
                    aid.log_warn(
                        1,
                        """
                        Phase patching did not remove all unstable roots in
                        a single pass. It is not programmed to accommodate this
                        failure mode. Unstable zeros (or poles) may be wrong.
                        """,
                    )
                check_second_pass = False

            phase_fix_about = Bunch()
            phase_fix_about.shift_saves = shift_saves
            phase_fix_about.fitter_X = fitter_X.copy()
            phase_fixes.append(phase_fix_about)

    def investigate_phase_patch_plot(self, fname=None, xscale="log_zoom", **kwargs):
        """
        Generates a plot of the mag normalized fit. Returns a Bunch object which
        stores the axes and figure of the plot.
        """
        from .. import plots

        axB = plots.plot_fitter_flag(
            fitter=phase_fixes[0].fitter_X, xscale=xscale, **kwargs
        )
        for pf in phase_fixes[1:]:
            axB.plot_fit(pf.fitter_X)
        if fname is not None:
            axB.save(fname)
        return axB

    aid.investigations["investigate_phase_patch_plot"] = investigate_phase_patch_plot

    def investigate_phase_patch_stats_plot(resaid, fname=None, xscale="log_zoom"):
        """
        Boy does this plot need documentation.
        """
        # TODO, put into a function for the annotater
        from wield.iirrational.utilities.mpl import mplfigB

        axB = mplfigB(Ncols=2, Nrows=2)
        for shift in shift_saves:
            axB.ax0_0.plot(shift.f_local, shift.shift.real)
            axB.ax1_0.plot(shift.f_local, shift.shift.imag)
            axB.ax0_1.plot(shift.f_local, shift.shift_err)
        axB.ax0_0.set_xscale(xscale)
        axB.ax1_0.set_xscale(xscale)
        axB.ax0_1.set_xscale(xscale)
        axB.finalize()
        if fname is not None:
            axB.save(fname)
        return axB

    aid.investigations[
        "investigate_phase_patch_stats_plot"
    ] = investigate_phase_patch_stats_plot

    if len(fitter_X.num_codings) > 10:
        aid.log_warn(
            3,
            (
                """
            Phase patching found an excessive number of phase loops. Rational fitting probably didn't perform well... Skipping phase patching.
            Check --mode=rational to see how bad that fit was. Use --poles and --zeros suggestions to improve it.
            """
            ),
        )
        return aid.fitter

    extra_p_c = []
    extra_z_c = []
    for coding in fitter_X.num_codings:
        i = coding.F_Hz
        r = coding.BW_Hz
        if r < 0:
            if aid.hint("never_unstable_poles"):
                aid.log_warn(
                    3,
                    (
                        """
                    High Confidence that an unstable pole exists at {}+{}i Hz!
                    Not adding it to the filter, due to 'never_unstable_poles'
                    """
                    ).format(-r, i),
                )
                continue
            else:
                aid.log_warn(
                    3,
                    (
                        """
                    High Confidence that an unstable pole exists at {}+{}i Hz!
                    Adding it to the filter. Prevent this with 'never_unstable_poles'
                    """
                    ).format(-r, i),
                )
        else:
            if aid.hint("never_unstable_zeros"):
                aid.log_warn(
                    8,
                    (
                        """
                    High Confidence that an unstable zero exists at {}+{}i Hz
                    Not adding it to the filter, due to 'never_unstable_zeros'
                    """
                    ).format(+r, i),
                )
                continue
            else:
                aid.log_warn(
                    5,
                    (
                        """
                    High Confidence that an unstable zero exists at {}+{}i Hz
                    Adding it to the filter. Prevent this with 'never_unstable_zeros'
                    """
                    ).format(+r, i),
                )
        extra_p_c.append(-r + 1j * i)
        extra_z_c.append(+r + 1j * i)

    extra_p_r = []
    extra_z_r = []
    ord_diff = -int(4 * fitter_X.delay_s * fitter_X.F_max_Hz)
    reldeg_max = aid.hint("relative_degree_max")
    reldeg_min = aid.hint("relative_degree_min")
    old_order = aid.fitter_orders().reldeg
    new_order = old_order + ord_diff
    if ord_diff < 0:
        if reldeg_min is not None and new_order < reldeg_min:
            ord_diff = old_order - reldeg_min
        extra_p_r = [-aid.fitter.F_max_Hz] * -ord_diff
    else:
        if reldeg_max is not None and new_order > reldeg_max:
            ord_diff = reldeg_max - old_order
        extra_z_r = [-aid.fitter.F_max_Hz] * ord_diff
    gain_adjust = aid.fitter.F_max_Hz ** ord_diff

    # TODO add restrictions on the order increase-decrease
    fitter_new = aid.fitter.regenerate(
        ZPKrep=aid.fitter.ZPKrep,
        z_c=list(fitter_mag.zeros.c) + extra_z_c,
        p_c=list(fitter_mag.poles.c) + extra_p_c,
        z_r=list(fitter_mag.zeros.r) + extra_z_r,
        p_r=list(fitter_mag.poles.r) + extra_p_r,
    )
    fitter_new.gain = fitter_mag.gain * gain_adjust
    # TODO make this set to the specified delay
    fitter_new.delay_s = 0

    with fitter_new.with_codings_only([fitter_new.gain_coding]):
        fitter_new.optimize()
    algorithms.optimize_anneal(aid, fitter_new)
    aid.fitter_update(
        fitter_new,
        representative=True,
    )
    return fitter_new



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from .rational_bases import (
    DataFilterBase,
)

from .rational_disc import RationalDiscFilter
from .rational_cheby import ChebychevFilter



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest
import numpy as np

import os.path as path

from wield.iirrational import v2
from wield.iirrational.v2 import testing

from wield.iirrational.testing.utilities import (
    sign_validate_and_plot_hint,
)

try:
    from IIRrational_test_data import IIRrational_data_dev
except ImportError:
    module_import_skip = True
else:
    module_import_skip = False


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_simple1(request, browser, plotsections, plot_verbosity):
    data_name = "PUM_long"
    data = IIRrational_data_dev(
        data_name,
        instance_num=1,
        set_num=1,
    )
    out = v2.data2filter(
        xfer=data.data,
        F_Hz=data.F_Hz,
        SNR=data.SNR,
        order_initial=10,
        delay_s=0,
        F_nyquist_Hz=None,
        poles=(-100, -50),
        zeros=(-10, -50),
        mode="fit",
        delay_s_max=0.01,
        h_infinity=0.99,
        h_infinity_deweight=0.1,
        # hints = [
        #    testing.validate_plot_log(__file__, request),
        # ],
    )

    print(out.as_scipy_signal_ZPKsw())
    print(out.as_foton_ZPKsf())
    print(out.as_foton_str_ZPKsf())
    print(out.as_foton_ZPKsn())
    print(out.as_foton_str_ZPKsn())
    return


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_simple1_vec(request, browser, plotsections, plot_verbosity):
    data_name = "PUM_long"
    data = IIRrational_data_dev(
        data_name,
        instance_num=1,
        set_num=1,
    )
    out = v2.data2filter(
        xfer=data.data,
        F_Hz=data.F_Hz,
        SNR=data.SNR,
        order_initial=10,
        delay_s=0,
        F_nyquist_Hz=None,
        poles=(-100, -50),
        zeros=(-10, -50),
        mode="fit",
        delay_s_max=0.01,
        h_infinity=abs(1 / (1 + 1j * data.F_Hz / 1)),
        # hints = [
        #    testing.validate_plot_log(__file__, request),
        # ],
    )

    print(out.as_scipy_signal_ZPKsw())
    print(out.as_foton_ZPKsf())
    print(out.as_foton_str_ZPKsf())
    print(out.as_foton_ZPKsn())
    print(out.as_foton_str_ZPKsn())
    return


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_simple1_callable(request, browser, plotsections, plot_verbosity):
    data_name = "PUM_long"
    data = IIRrational_data_dev(
        data_name,
        instance_num=1,
        set_num=1,
    )
    out = v2.data2filter(
        xfer=data.data,
        F_Hz=data.F_Hz,
        SNR=data.SNR,
        order_initial=10,
        delay_s=0,
        F_nyquist_Hz=None,
        poles=(-100, -50),
        zeros=(-10, -50),
        mode="fit",
        delay_s_max=0.01,
        h_infinity=lambda L: abs(1 / (1 + 1j * np.linspace(0, 100, L))),
        # hints = [
        #    testing.validate_plot_log(__file__, request),
        # ],
    )

    print(out.as_scipy_signal_ZPKsw())
    print(out.as_foton_ZPKsf())
    print(out.as_foton_str_ZPKsf())
    print(out.as_foton_ZPKsn())
    print(out.as_foton_str_ZPKsn())
    return



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

import os.path as path

from wield.iirrational.testing import IIRrational_data
from wield.iirrational import v1


def test_simple1(browser, plotsections, plot_verbosity):
    # this opens a live plot, which we do not want, should incorporate that
    # functionality into plot_on_assert
    return
    data_name = "simple2"
    out = v1.data2filter(
        IIRrational_data(
            data_name,
            instance_num=1,
            set_num=2,
        ),
        order_initial=10,
        delay_s=0,
    )

    import matplotlib.pyplot as plt
    from wield.iirrational import plots

    axB = plots.plot_fitter_flag(
        out.fitter,
        # don't use jpg as matplotlib can't write it out of the box (needs pillow)
        fname=path.join(path.split(__file__)[0], "out-simple/test.png"),
    )
    plt.close(axB.fig)
    plt.show()
    return


if __name__ == "__main__":
    test_simple1(False, None, 10)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from ..codings_cmn import (
    CodingType,
    # Ipi,
    # I2pi
)

# import scipy.linalg


class CodingnlFnlBW(CodingType):
    N_parameters = 2
    unstable = False
    p_nlF_Hz = 0
    p_nl_BW = 0
    min_BW_Hz = 0
    F_cutoff_Hz = None

    def setup(self):
        self.N_parameters = 2

    def update(self, A=None, B=None):
        nlF_Hz, nl_BW = A, B
        self.p_nlF_Hz = nlF_Hz
        self.p_nl_BW = nl_BW

    def reduce(self):
        return [
            self.p_nlF_Hz,
            self.p_nl_BW,
        ]

    def option_set(self, minimum_BW_Hz=None, **kwargs):
        super(CodingnlFnlBW, self).option_set(**kwargs)
        if minimum_BW_Hz is not None:
            self.min_BW_Hz = minimum_BW_Hz
        # TODO, should modify p_nl_BW for the new min_BW_Hz
        return

    def check_dist_limit(self, F_Hz=None, thresh=1):
        if self.sys.distance_limit_auto >= thresh:
            if F_Hz is None:
                self.min_BW_Hz, D = self.sys.distance_limit(
                    self.F_Hz, with_derivative=True
                )
            else:
                self.min_BW_Hz, D = self.sys.distance_limit(F_Hz, with_derivative=True)
            return D
        return 0

    @property
    def gain_effect(self):
        if not self.unstable:
            return 1
        else:
            return 1

    def update_roots(self, r1):
        """ """
        # TODO should honor min_BW_Hz
        F_Hz = abs(r1.imag)
        self.check_dist_limit(F_Hz=F_Hz, thresh=1)

        ret = True
        self.F_cutoff_Hz = self.sys.F_cutoff_Hz
        relF_Hz = F_Hz / (self.F_cutoff_Hz)
        if relF_Hz >= 1:
            relF_Hz = 0.999
            ret = False
        elif relF_Hz <= -1:
            relF_Hz = 0.001
            ret = False
        self.p_nlF_Hz = relF_Hz / (1 - relF_Hz ** 2) ** 0.5

        if r1.real <= 0:
            self.unstable = False
            r1p = -r1.real
        else:
            self.unstable = True
            r1p = r1.real
        if r1p > 2 * self.min_BW_Hz:
            self.p_nl_BW = (r1p - self.min_BW_Hz) ** 0.5
        else:
            self.p_nl_BW = self.min_BW_Hz ** 0.5
            ret = False
        return ret

    @property
    def relF_Hz(self):
        val = self.p_nlF_Hz / (1.0 + self.p_nlF_Hz ** 2) ** 0.5
        return val

    @property
    def F_Hz(self):
        return self.F_cutoff_Hz * self.relF_Hz

    @property
    def BW_Hz(self):
        return self.min_BW_Hz + self.p_nl_BW ** 2

    def transfer(self):
        self.check_dist_limit(thresh=2)
        # frequency, logarithmic amplitude
        if not self.unstable:
            r = -self.BW_Hz
        else:
            r = self.BW_Hz
        i = self.F_Hz
        X = self.sys.Xsf_grid
        Xsq = self.sys.Xsf_grid_sq
        return (r * r + i * i) - 2 * X * r + Xsq

    def derivative(self):
        D_mBW = self.check_dist_limit(thresh=2)
        # real/imaginary part of root
        if not self.unstable:
            r = -self.BW_Hz
            DrDp = -2 * self.p_nl_BW
            D_mBW *= -1
        else:
            r = self.BW_Hz
            DrDp = 2 * self.p_nl_BW

        Fsqp5 = (1.0 + self.p_nlF_Hz ** 2) ** 0.5
        # relF_Hz   = (self.p_nlF_Hz / Fsqp5)
        DrelFHzDp = 1 / Fsqp5 - self.p_nlF_Hz ** 2 / (1.0 + self.p_nlF_Hz ** 2) ** 1.5

        i = self.F_Hz
        X = self.sys.Xsf_grid

        # print("R", r)
        # print('D_mBW', D_mBW, sys.distance_limit_auto)
        # print(sys.distance_limit(self.F_Hz, with_derivative = True))
        # print("A", 2 * i)
        # print("B", D_mBW * (2 * r - 2 * X))
        # assert(False)
        # TODO, make this work and test it..
        # D_mBW = 0

        # TODO, need to test with D_mBW
        return [
            DrelFHzDp * self.F_cutoff_Hz * (2 * i + D_mBW * (2 * r - 2 * X)),
            2 * DrDp * r - 2 * X * DrDp,
        ]

    def roots_c(self):
        if not self.unstable:
            r = -self.BW_Hz
        else:
            r = self.BW_Hz
        i = self.F_cutoff_Hz * abs(self.relF_Hz)
        # real/imaginary part of root
        return [r + 1j * i]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from os import path

from .version import (
    version,
    __version__,
)

from . import v1
from . import v2
from .file_io import save, load


def matlabpath():
    # import msurrogate
    # don't join them as it only seems to work on linux
    return ":".join(
        [
            # msurrogate.matlabpath(),
            path.abspath(path.split(__file__)[0]),
        ]
    )


__all__ = [
    "version",
    "__version__",
    "v1,"
    "v2",
    "save,"
    "load",
    "matlabpath,"
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np
from wield import declarative
from .roots_matching import nearest_pairs
from .. import representations


def Q_rank_calc(z, p):
    if p.real == 0 or z.real == 0:
        if p.real == z.real:
            Q_rank = 0
        else:
            # TODO
            # should use the data spacing to regularize this case
            Q_rank = 1e3
    else:
        res_ratio = z.real / p.real
        Q_rank = abs(p - z) * (1 / (p.real) ** 2 + 1 / (z.real) ** 2) ** 0.5 + abs(
            res_ratio - 1 / res_ratio
        )
    return Q_rank


def Q_rank_single(r):
    Q_rank = abs(r.imag / r.real)
    return Q_rank


def order_reduce_zpk(
    zpk,
    Q_rank_cutoff=1e-5,
    Q_rank_cutoff_unstable=None,
    reduce_c=True,
    reduce_r=False,
    RBalgo=representations.root_bunch.RBalgo,
):
    zeros = RBalgo.expect_atleast(
        zpk[0], constraint=RBalgo.root_constraints.mirror_real
    )
    poles = RBalgo.expect_atleast(
        zpk[1], constraint=RBalgo.root_constraints.mirror_real
    )
    Pc = poles.c
    Zc = zeros.c
    Pr = poles.r
    Zr = zeros.r

    # print(len(Zc), len(Pc))
    if reduce_c:
        if Q_rank_cutoff_unstable:
            rpB = nearest_pairs(Zc[Zc.real > 0], Pc[Pc.real > 0])
            Zc = list(rpB.l1_remain) + list(Zc[Zc.real <= 0])
            Pc = list(rpB.l2_remain) + list(Pc[Pc.real <= 0])

            removed_rzp_list = []
            for z, p in rpB.r12_list:
                Q_rank = Q_rank_calc(p, z)
                # print(z, p, Q_rank)
                if Q_rank < Q_rank_cutoff_unstable:
                    removed_rzp_list.append((Q_rank, z, p))
                    continue
                Zc.append(z)
                Pc.append(p)
        rpB = nearest_pairs(Zc, Pc)
        Zc = list(rpB.l1_remain)
        Pc = list(rpB.l2_remain)

        removed_rzp_list = []
        for z, p in rpB.r12_list:
            Q_rank = Q_rank_calc(p, z)
            # print(z, p, Q_rank)
            if Q_rank < Q_rank_cutoff:
                removed_rzp_list.append((Q_rank, z, p))
                continue
            Zc.append(z)
            Pc.append(p)

    if reduce_r:
        if Q_rank_cutoff_unstable:
            rpB = nearest_pairs(Zr[Zr.real > 0], Pr[Pr.real > 0])
            Zr = list(rpB.l1_remain) + list(Zr[Zr.real <= 0])
            Pr = list(rpB.l2_remain) + list(Pr[Pr.real <= 0])

            removed_rzp_list = []
            for z, p in rpB.r12_list:
                Q_rank = Q_rank_calc(p, z)
                # print(z, p, Q_rank)
                if Q_rank < Q_rank_cutoff_unstable:
                    removed_rzp_list.append((Q_rank, z, p))
                    continue
                Zr.append(z)
                Pr.append(p)
        rpB = nearest_pairs(Zr, Pr)
        Zr = list(rpB.l1_remain)
        Pr = list(rpB.l2_remain)

        removed_rzp_list = []
        for z, p in rpB.r12_list:
            Q_rank = Q_rank_calc(p, z)
            # print(z, p, Q_rank)
            if Q_rank < Q_rank_cutoff:
                removed_rzp_list.append((Q_rank, z, p))
                continue
            Zr.append(z)
            Pr.append(p)

    return (
        np.concatenate([Zr, Zc, np.conjugate(Zc)]),
        np.concatenate([Pr, Pc, np.conjugate(Pc)]),
        zpk[2],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
from numpy.distutils.system_info import get_info, NotFoundError

from numpy.distutils.misc_util import Configuration


def configuration(parent_package="", top_path=None):
    lapack_opt = get_info("lapack_opt")
    config = Configuration("lapack", parent_package=parent_package, top_path=top_path)

    if not lapack_opt:
        raise NotFoundError("no lapack/blas resources found")

    config.add_extension(
        name="lapack_svd",
        sources=["lapack_svd.pyf.src"],
        extra_info=lapack_opt,
    )
    return config


if __name__ == "__main__":
    from setuptools import setup
    from numpy.distutils.core import setup

    setup(
        description="LAPACK Wrappers",
        author="Lee McCuller",
        author_email="Lee.McCuller@gmail.com",
        **configuration(top_path="").todict()
    )



# -*- coding: utf-8 -*-
"""
This code originally from module scipy.optimize.optimize
modified slightly to reduce dependencies.

SciPy license
Copyright © 2001, 2002 Enthought, Inc.
All rights reserved.

Copyright © 2003-2013 SciPy Developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer. Redistributions in binary
    form must reproduce the above copyright notice, this list of conditions and
    the following disclaimer in the documentation and/or other materials
    provided with the distribution. Neither the name of Enthought nor the names
    of the SciPy Developers may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


from math import copysign

import numpy as np
from numpy.linalg import norm

from scipy.linalg import cho_factor, cho_solve, LinAlgError
from scipy.sparse import issparse
from scipy.sparse.linalg import LinearOperator, aslinearoperator


EPS = np.finfo(float).eps


# Functions related to a trust-region problem.


def intersect_trust_region(x, s, Delta):
    """Find the intersection of a line with the boundary of a trust region.

    This function solves the quadratic equation with respect to t
    ||(x + s*t)||**2 = Delta**2.

    Returns
    -------
    t_neg, t_pos : tuple of float
        Negative and positive roots.

    Raises
    ------
    ValueError
        If `s` is zero or `x` is not within the trust region.
    """
    a = np.dot(s, s)
    if a == 0:
        raise ValueError("`s` is zero.")

    b = np.dot(x, s)

    c = np.dot(x, x) - Delta ** 2
    if c > 0:
        raise ValueError("`x` is not within the trust region.")

    d = np.sqrt(b * b - a * c)  # Root from one fourth of the discriminant.

    # Computations below avoid loss of significance, see "Numerical Recipes".
    q = -(b + copysign(d, b))
    t1 = q / a
    t2 = c / q

    if t1 < t2:
        return t1, t2
    else:
        return t2, t1


def solve_lsq_trust_region(
    n, m, uf, s, V, Delta, initial_alpha=None, rtol=0.01, max_iter=10
):
    """Solve a trust-region problem arising in least-squares minimization.

    This function implements a method described by J. J. More [1]_ and used
    in MINPACK, but it relies on a single SVD of Jacobian instead of series
    of Cholesky decompositions. Before running this function, compute:
    ``U, s, VT = svd(J, full_matrices=False)``.

    Parameters
    ----------
    n : int
        Number of variables.
    m : int
        Number of residuals.
    uf : ndarray
        Computed as U.T.dot(f).
    s : ndarray
        Singular values of J.
    V : ndarray
        Transpose of VT.
    Delta : float
        Radius of a trust region.
    initial_alpha : float, optional
        Initial guess for alpha, which might be available from a previous
        iteration. If None, determined automatically.
    rtol : float, optional
        Stopping tolerance for the root-finding procedure. Namely, the
        solution ``p`` will satisfy ``abs(norm(p) - Delta) < rtol * Delta``.
    max_iter : int, optional
        Maximum allowed number of iterations for the root-finding procedure.

    Returns
    -------
    p : ndarray, shape (n,)
        Found solution of a trust-region problem.
    alpha : float
        Positive value such that (J.T*J + alpha*I)*p = -J.T*f.
        Sometimes called Levenberg-Marquardt parameter.
    n_iter : int
        Number of iterations made by root-finding procedure. Zero means
        that Gauss-Newton step was selected as the solution.

    References
    ----------
    .. [1] More, J. J., "The Levenberg-Marquardt Algorithm: Implementation
           and Theory," Numerical Analysis, ed. G. A. Watson, Lecture Notes
           in Mathematics 630, Springer Verlag, pp. 105-116, 1977.
    """

    def phi_and_derivative(alpha, suf, s, Delta):
        """Function of which to find zero.

        It is defined as "norm of regularized (by alpha) least-squares
        solution minus `Delta`". Refer to [1]_.
        """
        denom = s ** 2 + alpha
        p_norm = norm(suf / denom)
        phi = p_norm - Delta
        phi_prime = -np.sum(suf ** 2 / denom ** 3) / p_norm
        return phi, phi_prime

    suf = s * uf

    # Check if J has full rank and try Gauss-Newton step.
    if m >= n:
        threshold = EPS * m * s[0]
        full_rank = s[-1] > threshold
    else:
        full_rank = False

    if full_rank:
        p = -V.dot(uf / s)
        if norm(p) <= Delta:
            return p, 0.0, 0

    alpha_upper = norm(suf) / Delta

    if full_rank:
        phi, phi_prime = phi_and_derivative(0.0, suf, s, Delta)
        alpha_lower = -phi / phi_prime
    else:
        alpha_lower = 0.0

    if initial_alpha is None or not full_rank and initial_alpha == 0:
        alpha = max(0.001 * alpha_upper, (alpha_lower * alpha_upper) ** 0.5)
    else:
        alpha = initial_alpha

    for it in range(max_iter):
        if alpha < alpha_lower or alpha > alpha_upper:
            alpha = max(0.001 * alpha_upper, (alpha_lower * alpha_upper) ** 0.5)

        phi, phi_prime = phi_and_derivative(alpha, suf, s, Delta)

        if phi < 0:
            alpha_upper = alpha

        ratio = phi / phi_prime
        alpha_lower = max(alpha_lower, alpha - ratio)
        alpha -= (phi + Delta) * ratio / Delta

        if np.abs(phi) < rtol * Delta:
            break

    p = -V.dot(suf / (s ** 2 + alpha))

    # Make the norm of p equal to Delta, p is changed only slightly during
    # this. It is done to prevent p lie outside the trust region (which can
    # cause problems later).
    p *= Delta / norm(p)

    return p, alpha, it + 1


def solve_trust_region_2d(B, g, Delta):
    """Solve a general trust-region problem in 2 dimensions.

    The problem is reformulated as a 4-th order algebraic equation,
    the solution of which is found by numpy.roots.

    Parameters
    ----------
    B : ndarray, shape (2, 2)
        Symmetric matrix, defines a quadratic term of the function.
    g : ndarray, shape (2,)
        Defines a linear term of the function.
    Delta : float
        Radius of a trust region.

    Returns
    -------
    p : ndarray, shape (2,)
        Found solution.
    newton_step : bool
        Whether the returned solution is the Newton step which lies within
        the trust region.
    """
    try:
        R, lower = cho_factor(B)
        p = -cho_solve((R, lower), g)
        if np.dot(p, p) <= Delta ** 2:
            return p, True
    except LinAlgError:
        pass

    a = B[0, 0] * Delta ** 2
    b = B[0, 1] * Delta ** 2
    c = B[1, 1] * Delta ** 2

    d = g[0] * Delta
    f = g[1] * Delta

    coeffs = np.array([-b + d, 2 * (a - c + f), 6 * b, 2 * (-a + c + f), -b - d])
    t = np.roots(coeffs)  # Can handle leading zeros.
    t = np.real(t[np.isreal(t)])

    p = Delta * np.vstack((2 * t / (1 + t ** 2), (1 - t ** 2) / (1 + t ** 2)))
    value = 0.5 * np.sum(p * B.dot(p), axis=0) + np.dot(g, p)
    i = np.argmin(value)
    p = p[:, i]

    return p, False


def update_tr_radius(
    Delta, actual_reduction, predicted_reduction, step_norm, bound_hit
):
    """Update the radius of a trust region based on the cost reduction.

    Returns
    -------
    Delta : float
        New radius.
    ratio : float
        Ratio between actual and predicted reductions. Zero if predicted
        reduction is zero.
    """
    if predicted_reduction > 0:
        ratio = actual_reduction / predicted_reduction
    else:
        ratio = 0

    if ratio < 0.25:
        Delta = 0.25 * step_norm
    elif ratio > 0.75 and bound_hit:
        Delta *= 2.0

    return Delta, ratio


# Construction and minimization of quadratic functions.


def build_quadratic_1d(J, g, s, diag=None, s0=None):
    """Parameterize a multivariate quadratic function along a line.

    The resulting univariate quadratic function is given as follows:
    ::
        f(t) = 0.5 * (s0 + s*t).T * (J.T*J + diag) * (s0 + s*t) +
               g.T * (s0 + s*t)

    Parameters
    ----------
    J : ndarray, sparse matrix or LinearOperator shape (m, n)
        Jacobian matrix, affects the quadratic term.
    g : ndarray, shape (n,)
        Gradient, defines the linear term.
    s : ndarray, shape (n,)
        Direction vector of a line.
    diag : None or ndarray with shape (n,), optional
        Addition diagonal part, affects the quadratic term.
        If None, assumed to be 0.
    s0 : None or ndarray with shape (n,), optional
        Initial point. If None, assumed to be 0.

    Returns
    -------
    a : float
        Coefficient for t**2.
    b : float
        Coefficient for t.
    c : float
        Free term. Returned only if `s0` is provided.
    """
    v = J.dot(s)
    a = np.dot(v, v)
    if diag is not None:
        a += np.dot(s * diag, s)
    a *= 0.5

    b = np.dot(g, s)

    if s0 is not None:
        u = J.dot(s0)
        b += np.dot(u, v)
        c = 0.5 * np.dot(u, u) + np.dot(g, s0)
        if diag is not None:
            b += np.dot(s0 * diag, s)
            c += 0.5 * np.dot(s0 * diag, s0)
        return a, b, c
    else:
        return a, b


def minimize_quadratic_1d(a, b, lb, ub, c=0):
    """Minimize a 1-d quadratic function subject to bounds.

    The free term `c` is 0 by default. Bounds must be finite.

    Returns
    -------
    t : float
        Minimum point.
    y : float
        Minimum value.
    """
    t = [lb, ub]
    if a != 0:
        extremum = -0.5 * b / a
        if lb < extremum < ub:
            t.append(extremum)
    t = np.asarray(t)
    y = a * t ** 2 + b * t + c
    min_index = np.argmin(y)
    return t[min_index], y[min_index]


def evaluate_quadratic(J, g, s, diag=None):
    """Compute values of a quadratic function arising in least squares.

    The function is 0.5 * s.T * (J.T * J + diag) * s + g.T * s.

    Parameters
    ----------
    J : ndarray, sparse matrix or LinearOperator, shape (m, n)
        Jacobian matrix, affects the quadratic term.
    g : ndarray, shape (n,)
        Gradient, defines the linear term.
    s : ndarray, shape (k, n) or (n,)
        Array containing steps as rows.
    diag : ndarray, shape (n,), optional
        Addition diagonal part, affects the quadratic term.
        If None, assumed to be 0.

    Returns
    -------
    values : ndarray with shape (k,) or float
        Values of the function. If `s` was 2-dimensional then ndarray is
        returned, otherwise float is returned.
    """
    if s.ndim == 1:
        Js = J.dot(s)
        q = np.dot(Js, Js)
        if diag is not None:
            q += np.dot(s * diag, s)
    else:
        Js = J.dot(s.T)
        q = np.sum(Js ** 2, axis=0)
        if diag is not None:
            q += np.sum(diag * s ** 2, axis=1)

    l = np.dot(s, g)

    return 0.5 * q + l


# Utility functions to work with bound constraints.


def in_bounds(x, lb, ub):
    """Check if a point lies within bounds."""
    return np.all((x >= lb) & (x <= ub))


def step_size_to_bound(x, s, lb, ub):
    """Compute a min_step size required to reach a bound.

    The function computes a positive scalar t, such that x + s * t is on
    the bound.

    Returns
    -------
    step : float
        Computed step. Non-negative value.
    hits : ndarray of int with shape of x
        Each element indicates whether a corresponding variable reaches the
        bound:

             *  0 - the bound was not hit.
             * -1 - the lower bound was hit.
             *  1 - the upper bound was hit.
    """
    non_zero = np.nonzero(s)
    s_non_zero = s[non_zero]
    steps = np.empty_like(x)
    steps.fill(np.inf)
    with np.errstate(over="ignore"):
        steps[non_zero] = np.maximum(
            (lb - x)[non_zero] / s_non_zero, (ub - x)[non_zero] / s_non_zero
        )
    min_step = np.min(steps)
    return min_step, np.equal(steps, min_step) * np.sign(s).astype(int)


def find_active_constraints(x, lb, ub, rtol=1e-10):
    """Determine which constraints are active in a given point.

    The threshold is computed using `rtol` and the absolute value of the
    closest bound.

    Returns
    -------
    active : ndarray of int with shape of x
        Each component shows whether the corresponding constraint is active:

             *  0 - a constraint is not active.
             * -1 - a lower bound is active.
             *  1 - a upper bound is active.
    """
    active = np.zeros_like(x, dtype=int)

    if rtol == 0:
        active[x <= lb] = -1
        active[x >= ub] = 1
        return active

    lower_dist = x - lb
    upper_dist = ub - x

    lower_threshold = rtol * np.maximum(1, np.abs(lb))
    upper_threshold = rtol * np.maximum(1, np.abs(ub))

    lower_active = np.isfinite(lb) & (
        lower_dist <= np.minimum(upper_dist, lower_threshold)
    )
    active[lower_active] = -1

    upper_active = np.isfinite(ub) & (
        upper_dist <= np.minimum(lower_dist, upper_threshold)
    )
    active[upper_active] = 1

    return active


def make_strictly_feasible(x, lb, ub, rstep=1e-10):
    """Shift a point to the interior of a feasible region.

    Each element of the returned vector is at least at a relative distance
    `rstep` from the closest bound. If ``rstep=0`` then `np.nextafter` is used.
    """
    x_new = x.copy()

    active = find_active_constraints(x, lb, ub, rstep)
    lower_mask = np.equal(active, -1)
    upper_mask = np.equal(active, 1)

    if rstep == 0:
        x_new[lower_mask] = np.nextafter(lb[lower_mask], ub[lower_mask])
        x_new[upper_mask] = np.nextafter(ub[upper_mask], lb[upper_mask])
    else:
        x_new[lower_mask] = lb[lower_mask] + rstep * np.maximum(
            1, np.abs(lb[lower_mask])
        )
        x_new[upper_mask] = ub[upper_mask] - rstep * np.maximum(
            1, np.abs(ub[upper_mask])
        )

    tight_bounds = (x_new < lb) | (x_new > ub)
    x_new[tight_bounds] = 0.5 * (lb[tight_bounds] + ub[tight_bounds])

    return x_new


def CL_scaling_vector(x, g, lb, ub):
    """Compute Coleman-Li scaling vector and its derivatives.

    Components of a vector v are defined as follows:
    ::
               | ub[i] - x[i], if g[i] < 0 and ub[i] < np.inf
        v[i] = | x[i] - lb[i], if g[i] > 0 and lb[i] > -np.inf
               | 1,           otherwise

    According to this definition v[i] >= 0 for all i. It differs from the
    definition in paper [1]_ (eq. (2.2)), where the absolute value of v is
    used. Both definitions are equivalent down the line.
    Derivatives of v with respect to x take value 1, -1 or 0 depending on a
    case.

    Returns
    -------
    v : ndarray with shape of x
        Scaling vector.
    dv : ndarray with shape of x
        Derivatives of v[i] with respect to x[i], diagonal elements of v's
        Jacobian.

    References
    ----------
    .. [1] M.A. Branch, T.F. Coleman, and Y. Li, "A Subspace, Interior,
           and Conjugate Gradient Method for Large-Scale Bound-Constrained
           Minimization Problems," SIAM Journal on Scientific Computing,
           Vol. 21, Number 1, pp 1-23, 1999.
    """
    v = np.ones_like(x)
    dv = np.zeros_like(x)

    mask = (g < 0) & np.isfinite(ub)
    v[mask] = ub[mask] - x[mask]
    dv[mask] = -1

    mask = (g > 0) & np.isfinite(lb)
    v[mask] = x[mask] - lb[mask]
    dv[mask] = 1

    return v, dv


def reflective_transformation(y, lb, ub):
    """Compute reflective transformation and its gradient."""
    if in_bounds(y, lb, ub):
        return y, np.ones_like(y)

    lb_finite = np.isfinite(lb)
    ub_finite = np.isfinite(ub)

    x = y.copy()
    g_negative = np.zeros_like(y, dtype=bool)

    mask = lb_finite & ~ub_finite
    x[mask] = np.maximum(y[mask], 2 * lb[mask] - y[mask])
    g_negative[mask] = y[mask] < lb[mask]

    mask = ~lb_finite & ub_finite
    x[mask] = np.minimum(y[mask], 2 * ub[mask] - y[mask])
    g_negative[mask] = y[mask] > ub[mask]

    mask = lb_finite & ub_finite
    d = ub - lb
    t = np.remainder(y[mask] - lb[mask], 2 * d[mask])
    x[mask] = lb[mask] + np.minimum(t, 2 * d[mask] - t)
    g_negative[mask] = t > d[mask]

    g = np.ones_like(y)
    g[g_negative] = -1

    return x, g


# Functions to display algorithm's progress.


def print_header_nonlinear():
    print(
        "{0:^15}{1:^15}{2:^15}{3:^15}{4:^15}{5:^15}".format(
            "Iteration",
            "Total nfev",
            "Cost",
            "Cost reduction",
            "Step norm",
            "Optimality",
        )
    )


def print_iteration_nonlinear(
    iteration, nfev, cost, cost_reduction, step_norm, optimality
):
    if cost_reduction is None:
        cost_reduction = " " * 15
    else:
        cost_reduction = "{0:^15.2e}".format(cost_reduction)

    if step_norm is None:
        step_norm = " " * 15
    else:
        step_norm = "{0:^15.2e}".format(step_norm)

    print(
        "{0:^15}{1:^15}{2:^15.4e}{3}{4}{5:^15.2e}".format(
            iteration, nfev, cost, cost_reduction, step_norm, optimality
        )
    )


def print_header_linear():
    print(
        "{0:^15}{1:^15}{2:^15}{3:^15}{4:^15}".format(
            "Iteration", "Cost", "Cost reduction", "Step norm", "Optimality"
        )
    )


def print_iteration_linear(iteration, cost, cost_reduction, step_norm, optimality):
    if cost_reduction is None:
        cost_reduction = " " * 15
    else:
        cost_reduction = "{0:^15.2e}".format(cost_reduction)

    if step_norm is None:
        step_norm = " " * 15
    else:
        step_norm = "{0:^15.2e}".format(step_norm)

    print(
        "{0:^15}{1:^15.4e}{2}{3}{4:^15.2e}".format(
            iteration, cost, cost_reduction, step_norm, optimality
        )
    )


# Simple helper functions.


def compute_grad(J, f):
    """Compute gradient of the least-squares cost function."""
    if isinstance(J, LinearOperator):
        return J.rmatvec(f)
    else:
        return J.T.dot(f)


def compute_jac_scale(J, scale_inv_old=None):
    """Compute variables scale based on the Jacobian matrix."""
    if issparse(J):
        scale_inv = np.asarray(J.power(2).sum(axis=0)).ravel() ** 0.5
    else:
        scale_inv = np.sum(J ** 2, axis=0) ** 0.5

    if scale_inv_old is None:
        scale_inv[scale_inv == 0] = 1
    else:
        scale_inv = np.maximum(scale_inv, scale_inv_old)

    return 1 / scale_inv, scale_inv


def left_multiplied_operator(J, d):
    """Return diag(d) J as LinearOperator."""
    J = aslinearoperator(J)

    def matvec(x):
        return d * J.matvec(x)

    def matmat(X):
        return d * J.matmat(X)

    def rmatvec(x):
        return J.rmatvec(x.ravel() * d)

    return LinearOperator(J.shape, matvec=matvec, matmat=matmat, rmatvec=rmatvec)


def right_multiplied_operator(J, d):
    """Return J diag(d) as LinearOperator."""
    J = aslinearoperator(J)

    def matvec(x):
        return J.matvec(np.ravel(x) * d)

    def matmat(X):
        return J.matmat(X * d[:, np.newaxis])

    def rmatvec(x):
        return d * J.rmatvec(x)

    return LinearOperator(J.shape, matvec=matvec, matmat=matmat, rmatvec=rmatvec)


def regularized_lsq_operator(J, diag):
    """Return a matrix arising in regularized least squares as LinearOperator.

    The matrix is
        [ J ]
        [ D ]
    where D is diagonal matrix with elements from `diag`.
    """
    J = aslinearoperator(J)
    m, n = J.shape

    def matvec(x):
        return np.hstack((J.matvec(x), diag * x))

    def rmatvec(x):
        x1 = x[:m]
        x2 = x[m:]
        return J.rmatvec(x1) + diag * x2

    return LinearOperator((m + n, n), matvec=matvec, rmatvec=rmatvec)


def right_multiply(J, d, copy=True):
    """Compute J diag(d).

    If `copy` is False, `J` is modified in place (unless being LinearOperator).
    """
    if copy and not isinstance(J, LinearOperator):
        J = J.copy()

    if issparse(J):
        J.data *= d.take(J.indices, mode="clip")  # scikit-learn recipe.
    elif isinstance(J, LinearOperator):
        J = right_multiplied_operator(J, d)
    else:
        J *= d

    return J


def left_multiply(J, d, copy=True):
    """Compute diag(d) J.

    If `copy` is False, `J` is modified in place (unless being LinearOperator).
    """
    if copy and not isinstance(J, LinearOperator):
        J = J.copy()

    if issparse(J):
        J.data *= np.repeat(d, np.diff(J.indptr))  # scikit-learn recipe.
    elif isinstance(J, LinearOperator):
        J = left_multiplied_operator(J, d)
    else:
        J *= d[:, np.newaxis]

    return J


def check_termination(dF, F, dx_norm, x_norm, ratio, ftol, xtol):
    """Check termination condition for nonlinear least squares."""
    ftol_satisfied = dF < ftol * F and ratio > 0.25
    xtol_satisfied = dx_norm < xtol * (xtol + x_norm)

    if ftol_satisfied and xtol_satisfied:
        return 4
    elif ftol_satisfied:
        return 2
    elif xtol_satisfied:
        return 3
    else:
        return None


def scale_for_robust_loss_function(J, f, rho):
    """Scale Jacobian and residuals for a robust loss function.

    Arrays are modified in place.
    """
    J_scale = rho[1] + 2 * rho[2] * f ** 2
    J_scale[J_scale < EPS] = EPS
    J_scale **= 0.5

    f *= rho[1] / J_scale

    return left_multiply(J, J_scale, copy=False), f



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

import os.path as path

from wield.iirrational.testing import IIRrational_data
from wield.iirrational import v2
from wield.iirrational.v2.__main__ import main as v2_main
from wield.iirrational.v2 import testing

from wield.iirrational.testing.utilities import (
    sign_validate_and_plot_hint,
)


def test_argtell(request, browser, plotsections, plot_verbosity):
    dset = IIRrational_data(
        "simple1",
        instance_num=1,
        set_num=1,
    )

    res = v2.data2filter(
        F_Hz=dset.F_Hz,
        SNR=dset.SNR,
        data=dset.data,
        order_initial=10,
        delay_s=0,
        F_nyquist_Hz=None,
        mode="fit",
        delay_f=5,
        hints=[
            testing.validate_plot_log(__file__, request),
        ],
    )
    return


def test_prog_help(request, browser, plotsections, plot_verbosity):
    dset = IIRrational_data(
        "simple1",
        instance_num=1,
        set_num=1,
    )

    v2_main(["-h"])
    return


def test_prog(request, browser, plotsections, plot_verbosity):
    dset = IIRrational_data(
        "simple1",
        instance_num=1,
        set_num=1,
    )

    print(v2_main(["-z", "+1-1i", "+1+1J"]))
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import os.path as path
import subprocess


def pandoc_pdf(out_name):
    def exporter(md_name, folder):
        subprocess.call(["pandoc", "-o", out_name, md_name], cwd=folder)

    return exporter


def npm_markdown_pdf(out_name):
    def exporter(md_name, folder):
        subprocess.call(["markdown-pdf", md_name, "-o", out_name], cwd=folder)

    return exporter



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
Utilities to manipulate ZPK roots S to/from Z, and make transfer functions
"""


import numpy as np

from ..TFmath import match_SOS_pairs
from .. import representations


def filter2sortedZPK(filt):
    zpk = representations.asZPKTF(filt)
    z, p, k = zpk

    zzpp_pairs = match_SOS_pairs(
        zpk.zeros.r, zpk.zeros.c, zpk.poles.r, zpk.poles.c, F_nyquist_Hz=None
    )

    poles = []
    zeros = []

    for z1, z2, p1, p2 in zzpp_pairs:
        if z1 is not None:
            zeros.append(z1)
        if z2 is not None:
            zeros.append(z2)
        if p1 is not None:
            poles.append(p1)
        if p2 is not None:
            poles.append(p2)
    return zeros, poles, k


def filter2fotonZPK(
    filt,
    annotate_pairs=False,
    scale_gain=True,
    plane=None,
    zpk_output=False,
):
    if plane is None:
        plane = "f"
    zpk = representations.asZPKTF(filt)

    zzpp_pairs = match_SOS_pairs(
        zpk.zeros.r, zpk.zeros.c, zpk.poles.r, zpk.poles.c, F_nyquist_Hz=None
    )
    k = zpk.gain

    if scale_gain:
        if plane == "f":
            k = k / ((np.pi * 2) ** (len(zpk.zeros) - len(zpk.poles)))
        elif plane == "w":
            pass
        elif plane == "n":

            def root_gain(r1, r2):
                k_adj = 1
                if r1 is not None:
                    r1 = -r1.conjugate()
                    aval = abs(r1)
                    if aval > 1e-9:
                        k_adj = aval

                if r2 is not None:
                    aval = abs(r2)
                    if aval > 1e-9:
                        k_adj = k_adj * aval
                    r2 = -r2.conjugate()

                return r1, r2, k_adj

            zzpp_pairs_mod = []
            for idx, (z1, z2, p1, p2) in enumerate(zzpp_pairs):
                z1, z2, k_adj = root_gain(z1, z2)
                k *= k_adj
                p1, p2, k_adj = root_gain(p1, p2)
                k /= k_adj
                zzpp_pairs_mod.append((z1, z2, p1, p2))
            zzpp_pairs = zzpp_pairs_mod

    if zpk_output:
        zero_list = []
        pole_list = []

        def root_collect(r1, r2):
            ret = []
            if r1 is not None:
                ret = [r1]
            else:
                ret = []

            if r2 is not None:
                ret += [r2]
            return ret

        for idx, (z1, z2, p1, p2) in enumerate(zzpp_pairs):
            zero_list.extend(
                root_collect(
                    z1,
                    z2,
                )
            )
            pole_list.extend(
                root_collect(
                    p1,
                    p2,
                )
            )
        return zero_list, pole_list, k

    def matlab_complex_str(num):
        if num is None:
            return None
        if np.imag(num) == 0.0:
            return "{0}".format(num)
        else:
            r, i = np.real(num), np.imag(num)
            if i > 0:
                return "{0} + {1}*i".format(r, i)
            else:
                return "{0} - {1}*i".format(r, -i)

    if annotate_pairs:

        def root_strs(r1, r2, idx):
            s1 = matlab_complex_str(r1)
            s2 = matlab_complex_str(r2)
            if r2 is not None and r1 is not None:
                if len(s2) + len(s1) < 60:
                    return [
                        "{}; {};  % SOS {}AB".format(s1, s2, idx),
                    ]
                else:
                    return [
                        "{}; % SOS {}A".format(s1, idx),
                        "{}; % SOS {}B".format(s2, idx),
                    ]
            elif r1 is not None:
                return [
                    "{}; % SOS {}".format(s1, idx),
                ]
            elif r2 is not None:
                return [
                    "{}; % SOS {}".format(s2, idx),
                ]
            else:
                return []

    else:

        def root_strs(r1, r2, idx):
            s1 = matlab_complex_str(r1)
            s2 = matlab_complex_str(r2)
            if r2 is not None and r1 is not None:
                if len(s2) + len(s1) < 180:
                    return [
                        "{}; {};".format(s1, s2, idx),
                    ]
                else:
                    return [
                        "{};".format(s1, idx),
                        "{};".format(s2, idx),
                    ]
            elif r1 is not None:
                return [
                    "{};".format(s1, idx),
                ]
            elif r2 is not None:
                return [
                    "{};".format(s2, idx),
                ]
            else:
                return []

    def print_single(zzpp_pairs, gain):
        pole_list = []
        zero_list = []
        for idx, (z1, z2, p1, p2) in enumerate(zzpp_pairs):
            zero_list.extend(
                root_strs(
                    z1,
                    z2,
                    idx + 1,
                )
            )
            pole_list.extend(
                root_strs(
                    p1,
                    p2,
                    idx + 1,
                )
            )

        ZPK_template = 'ZPK([{zeros}],[{poles}], {gain}, "{plane}")'
        if len(zero_list) == 0:
            zstr = ""
        else:
            zstr = "\n  " + "\n  ".join(zero_list) + "\n"

        if len(pole_list) == 0:
            pstr = ""
        else:
            pstr = "\n  " + "\n  ".join(pole_list) + "\n"

        return ZPK_template.format(
            zeros=zstr,
            poles=pstr,
            gain=gain,
            plane=plane,
        )

    ZPKs = print_single(zzpp_pairs[:10], k)
    while True:
        zzpp_pairs = zzpp_pairs[10:]
        if not zzpp_pairs:
            break
        ZPKs = ZPKs + '\n' + print_single(zzpp_pairs[:10], 1)
    return ZPKs


'''
def filter2matlabZPK(cascade):
    def matlab_complex_str(num):
        if(np.imag(num) == 0.):
            return ("{0}".format(num))
        else:
            return ("{0} + {1}*i".format(np.real(num), np.imag(num)))

    cascade = cascade.transformed_nyquist(F_nyquist=None)
    ZPK_template = 'zpk(2*pi*[;...\n{zeros}],2*pi*[;...\n{poles}],...\n{gain})'
    pole_list = []
    zero_list = []
    g_fix = 1
    zpk_num_mismatch = 0

    for pole in cascade.poles:
        if pole is not None:
            pole_list.append(matlab_complex_str(pole))
            g_fix *= pole
            zpk_num_mismatch += 1

    for zero in cascade.zeros:
        if zero is not None:
            zero_list.append(matlab_complex_str(zero))
            g_fix *= zero
            zpk_num_mismatch += 1

    return ZPK_template.format(
        zeros=';...\n\t' + ';...\n\t'.join(zero_list) + ';...\n',
        poles=';...\n\t' + ';...\n\t'.join(pole_list) + ';...\n',
        gain=cascade.gain * abs(g_fix) * (2*np.pi)**zpk_num_mismatch
    )


def filter2fotonSOS(cascade):
    raise NotImplementedError()
    if cascade.F_nyquist is None:
        print("Warning: cascade needs to be converted to Z domain for foton sos representation")
    SOS_template = 'sos({gain}, [{coeffs}])'
    coeff_list = []
    for idx in range(cascade.num_sos):
        coeff_list.append(cascade.coeff('b1', idx, default = True))
        coeff_list.append(cascade.coeff('b2', idx, default = True))
        coeff_list.append(cascade.coeff('a1', idx, default = True))
        coeff_list.append(cascade.coeff('a2', idx, default = True))
    def none_to_0(f):
        if f is None:
            return 0
        return f
    coeff_list = [str(none_to_0(i)) for i in coeff_list]
    return SOS_template.format(gain=cascade.gain, coeffs=';'.join(coeff_list))


def filter2fotonZRoots(filter):
    """

    """
    raise NotImplementedError()
    def matlab_complex_str(num):
        if(np.imag(num) == 0.):
            return ("{0}".format(num))
        else:
            return ("{0} - {1}*i".format(np.real(num), np.imag(num)))

    cascade = cascade.transformed_nyquist(F_nyquist=None)
    ZPK_template = 'ZPK([{zeros}],[{poles}],\n{gain}, "f")'
    pole_list = []
    zero_list = []
    for pole in cascade.poles:
        pole = -pole.conjugate()
        pole_list.append(matlab_complex_str(pole))
    for zero in cascade.poles:
        zero = -zero.conjugate()
        pole_list.append(matlab_complex_str(zero))

    return ZPK_template.format(
        zeros='\n\t' + ';\n\t'.join(zero_list) + '\n',
        poles='\n\t' + ';\n\t'.join(pole_list) + '\n',
        gain=cascade.gain
    )
'''
